import Foundation
/*:
## **Aufgabe Krümel und Monster Prototyp 1**
*/

/**
Stellt den Gewinntext dar

- author: Albert Wiedemann
- version: 1.0
*/
class Gewonnen: Text
{

    /**
    Legt Position und Text fest
    */
    override init()
    {
        super.init()
        TextSetzen(text: "Du hast das Spiel gewonnen")
        PositionSetzen(x: 10, y: 20)
    }
}

/**
Stellt den Verlusttext dar

- author: Albert Wiedemann
- version: 1.0
*/
class Verloren: Text
{

    /**
     * Legt Position und Text fest
     */
    override init()
    {
        super.init()
        TextSetzen(text: "Du hast verloren")
        PositionSetzen(x: 500, y: 20)
    }
}


/**
Anzeige des Statustextes

- author: Albert Wiedemann
- version: 1.0
*/
class Status: Text
{

    /**
     * Legt die Position fest
     */
    override init()
    {
        super.init()
        PositionSetzen(x: 150, y: 15)
    }
    
    /**
	Trägt die Werte ein
	- parameters:
		- gefressen: Anzahl gefressener Krümel
		- gesamt: Gesamtzahl aller Krümel
		- leben: Anzahl der restlchen Leben
    */
    func AnzeigeSetzen(gefressen: Int, gesamt: Int, leben: Int)
    {
        TextSetzen(text: "Du hast \(gefressen) von \(gesamt) Krümeln gefressen und noch \(leben) Leben")
    }
}

/**
Verwaltet das Monster

- author: Albert Wiedemann
- version: 1.0
*/
class Monster: Figur
{
    /** Anzahl der restlichen Leben */
    private var leben: Int
    /** Mundstellung */
    private var mundIstZu: Bool
    /** Rahmenobjekt */
    private let rahmen: Rahmen

    /**
	Legt die Figur, die Mundstellung, die Größe und die Position fest.
	- parameters:
		- x: x-Position des Monsters
		- y: y-Position des Monsters
		- rahmen: das Rahmenobjekt
		- leben: Anzahl der Leben beim Start
    */
    init(x: Int, y: Int, rahmen: Rahmen, leben: Int)
    {
        self.leben = leben
        mundIstZu = true
        self.rahmen = rahmen
        super.init()
        PositionSetzen(x: x * rahmen.basisGröße + rahmen.basisGröße / 2, y: y * rahmen.basisGröße + rahmen.basisGröße / 2)
        GrößeSetzen(größe: rahmen.basisGröße - 2)
        Schließen(farbe: "gelb")
    }
    
    /**
    Stellt das gelb gefärbte Monster mit geschlossenem Mund dar.
    */
    func GeschlossenGelb()
    {
        EigeneFigurLöschen()
        FigurteilFestlegenEllipse(x: -50, y: -50, breite: 100, höhe: 100, farbe: "gelb")
        FigurteilFestlegenRechteck(x: 0, y: -2, breite: 50, höhe: 4, farbe: "schwarz")
        FigurteilFestlegenEllipse(x: 10, y: -40, breite: 20, höhe: 20, farbe: "schwarz")
    }

    /**
    Stellt das orange gefärbte Monster mit geschlossenem Mund dar.
    */
    func GeschlossenOrange()
    {
        EigeneFigurLöschen()
        FigurteilFestlegenEllipse(x: -50, y: -50, breite: 100, höhe: 100, farbe: "orange")
        FigurteilFestlegenRechteck(x: 0, y: -2, breite: 50, höhe: 4, farbe: "schwarz")
        FigurteilFestlegenEllipse(x: 10, y: -40, breite: 20, höhe: 20, farbe: "schwarz")
    }

    /**
    Stellt das rot gefärbte Monster mit geschlossenem Mund dar.
    */
    func GeschlossenRot()
    {
        EigeneFigurLöschen()
        FigurteilFestlegenEllipse(x: -50, y: -50, breite: 100, höhe: 100, farbe: "rot")
        FigurteilFestlegenRechteck(x: 0, y: -2, breite: 50, höhe: 4, farbe: "schwarz")
        FigurteilFestlegenEllipse(x: 10, y: -40, breite: 20, höhe: 20, farbe: "schwarz")
    }

    /**
    Stellt das grün gefärbte Monster mit geschlossenem Mund dar.
    */
    func GeschlossenGrün()
    {
        EigeneFigurLöschen()
        FigurteilFestlegenEllipse(x: -50, y: -50, breite: 100, höhe: 100, farbe: "grün")
        FigurteilFestlegenRechteck(x: 0, y: -2, breite: 50, höhe: 4, farbe: "schwarz")
        FigurteilFestlegenEllipse(x: 10, y: -40, breite: 20, höhe: 20, farbe: "schwarz")
    }

    /**
	Stellt das Monster in der angegebenen Farbe mit offenem Mund dar.
	- parameters:
		- farbe: die gewünschte Farbe
    */
    func Offen(farbe: String)
    {
        EigeneFigurLöschen()
        FigurteilFestlegenEllipse(x: -50, y: -50, breite: 100, höhe: 100, farbe: farbe)
        FigurteilFestlegenDreieck(x1: 0, y1: 0, x2: 50, y2: -10, x3: 50, y3: 10, farbe: "schwarz");
        FigurteilFestlegenEllipse(x: 10, y: -40, breite: 20, höhe: 20, farbe: "schwarz")
    }
    
    /**
	Stellt das in einer der Farben gelb, orange, rot bzw. grün mit geschlossenem Mund dar.
	- parameters:
		- farbe: die gewünschte Farbe
    */
    func Schließen(farbe: String)
    {
        if farbe == "gelb"
        {
            GeschlossenGelb()
        }
        if farbe == "orange"
        {
            GeschlossenOrange()
        }
        if farbe == "rot"
        {
            GeschlossenRot()
        }
        if farbe == "grün"
        {
            GeschlossenGrün()
        }
    }
    
    /**
	Stellt das in einer der Farben gelb, orange, rot bzw. grün mit offenem Mund dar.
	- parameters:
		- farbe: die gewünschte Farbe
    */
    func Öffnen(farbe: String)
    {
        if (farbe == "gelb") || (farbe == "orange") || (farbe == "rot") || (farbe == "grün")
        {
            Offen(farbe: farbe)
        }
    }
    
    /**
    Meldet die Anzahl der Leben
	- returns: Anzahl der Leben
    */
    func LebenGeben() -> Int
    {
        return leben
    }
    
    /**
    Reduziert die Zahl der Leben um 1
    */
    func LebenNehmen()
    {
        leben = leben - 1
        rahmen.StatusÄndern()
    }
    
    /**
    Die Aktionsmethode für gedrückte Sondertasten.
	- parameters:
		- taste: KeyCode der gedrückten Taste
    */
    override func SonderTasteGedrückt (taste: Int)
    {
        switch taste
        {
            case 123:
                Drehen(grad: 90)
            case 126:
                Gehen(länge: Double(rahmen.basisGröße));
                for h in rahmen.HindernisseGeben()
                {
                    if Berührt(objekt: h)
                    {
                        h.Reagieren(monster: self)
                    }
                }
            case 124:
                Drehen(grad: -90)
            case 125:
                Drehen(grad: 180)
            default:
                break
        }
    }
    
    /**
    Das Monster geht
    */
    override func AktionAusführen()
    {
        if mundIstZu
        {
            Öffnen(farbe: "gelb")
            mundIstZu = false
        }
        else
        {
            Schließen(farbe: "gelb")
            mundIstZu = true
        }
        if (XPositionGeben() == rahmen.basisGröße / 2) && (WinkelGeben() == 180)
        {
            Gehen(länge: Double(-27 * rahmen.basisGröße))
        }
        else if (XPositionGeben() == 27 * rahmen.basisGröße + rahmen.basisGröße / 2) && (WinkelGeben() == 0)
        {
            Gehen(länge: Double(-27 * rahmen.basisGröße))
        }
        else
        {
            Gehen(länge: Double(rahmen.basisGröße));
        }
        for h in rahmen.HindernisseGeben()
        {
            if Berührt(objekt: h)
            {
                h.Reagieren(monster: self)
            }
        }
    }
}

/**
Oberklasse für alle Hindernisse
Regelt die Größe und die Positionierung

- author: Albert Wiedemann
- version: 1.0
*/
class Hindernis: Figur
{
    /** Das Rahmenobjekt des Spiels */
    fileprivate let rahmen: Rahmen
    
    /**
	Der Konstruktor legt die Position fest
	- parameters:
		- x: x-Koordinate des Hindernisses
		- y: y-Koordinate des Hindernisses
		- rahmen: das Rahmenobjekt des Spiels
    */
    init(x: Int, y: Int, rahmen: Rahmen)
    {
        self.rahmen = rahmen
        super.init()
        PositionSetzen(x: x * rahmen.basisGröße + rahmen.basisGröße / 2, y: y * rahmen.basisGröße + rahmen.basisGröße / 2)
    }
    
    /**
	Das Hindernis kann auf des Monster reagieren
	- parameters:
		- monster: das Monster
    */
    func Reagieren(monster: Monster)
    {
        //Hier passiert nichts.
    }
}

/**
Beschreibt ein Rechteck der (grauen) Mauer

- author: Albert Wiedemann
- version: 1.0
*/
class MauerRechteck: Hindernis
{

    /**
	Der Konstruktor legt die Position fest
	- parameters:
		- x: x-Koordinate des Hindernisses
		- y: y-Koordinate des Hindernisses
		- rahmen: das Rahmenobjekt des Spiels
    */
    override init(x: Int, y: Int, rahmen: Rahmen)
    {
        super.init(x: x, y: y, rahmen: rahmen)
        EigeneFigurLöschen()
        FigurteilFestlegenRechteck(x: -50, y: -50, breite: 100, höhe: 100, farbe: "grau")
        GrößeSetzen(größe: rahmen.basisGröße)
    }
    
    /**
    Die Mauer befördert das Monster wieder zurück
     * @param monster das Monster
    */
    override func Reagieren(monster: Monster)
    {
        monster.Gehen(länge: -Double(rahmen.basisGröße))
    }
}

/**
Verwaltet einen Krümel

- author: Albert Wiedemann
- version: 1.0
*/
class Krümel: Hindernis
{

    /**
	Der Konstruktor legt die Position fest
	- parameters:
		- x: x-Koordinate des Hindernisses
		- y: y-Koordinate des Hindernisses
		- rahmen: das Rahmenobjekt des Spiels
    */
    override init(x: Int, y: Int, rahmen: Rahmen)
    {
        super.init(x: x, y: y, rahmen: rahmen)
        EigeneFigurLöschen()
        FigurteilFestlegenEllipse(x: -10, y: -10, breite: 20, höhe: 20, farbe: "orange")
        GrößeSetzen(größe: rahmen.basisGröße)
    }
    
    /**
	Das Krümel meldet dem Rahmen, dass es gefressen wurde
	- parameters:
		- monster: das Monster
    */
    override func Reagieren(monster: Monster)
    {
        rahmen.KrümelEntfernen(krümel: self)
        Entfernen()
    }
}

/**
Verfolgt das Monster bis zu einem Abstand von 10 und bewegt sich dann zufällig.

- author: Albert Wiedemann
- version: 1.0
*/
class Chaot: Hindernis
{
    /** x-Startposition merken */
    var xStart: Int
    /** y-Startposition merken */
    var yStart: Int
    /** Der Zufallsgenerator */
    var zufall: Zufall
    
    /**
	Der Konstruktor legt die Position fest
	- parameters:
		- x: x-Koordinate des Hindernisses
		- y: y-Koordinate des Hindernisses
		- rahmen: das Rahmenobjekt des Spiels
    */
    override init(x: Int, y: Int, rahmen: Rahmen)
    {
        xStart = x
        yStart = y
        zufall = Zufall()
        super.init(x: x, y: y, rahmen: rahmen)
        FigurErzeugen(farbe: "blau")
        GrößeSetzen(größe: rahmen.basisGröße)
    }
    
    /**
	Erzeugt das Figurenbild in der gewünschten Farbe
	- parameters:
		- farbe: die gewünschte Farbe
    */
    private func FigurErzeugen(farbe: String)
    {
        EigeneFigurLöschen()
        FigurteilFestlegenEllipse(x: -15, y: -50, breite: 30, höhe: 30, farbe: farbe)
        FigurteilFestlegenDreieck(x1: -25, y1: 45, x2: 0, y2: -50, x3: 25, y3: 45, farbe: farbe)
        FigurteilFestlegenRechteck(x: -15, y: -35, breite: 30, höhe: 80, farbe: farbe)
        FigurteilFestlegenEllipse(x: -25, y: 40, breite: 10, höhe: 10, farbe: farbe)
        FigurteilFestlegenEllipse(x: -15, y: 40, breite: 10, höhe: 10, farbe: farbe)
        FigurteilFestlegenEllipse(x: -5, y: 40, breite: 10, höhe: 10, farbe: farbe)
        FigurteilFestlegenEllipse(x: 5, y: 40, breite: 10, höhe: 10, farbe: farbe)
        FigurteilFestlegenEllipse(x: 15, y: 40, breite: 10, höhe: 10, farbe: farbe)
        FigurteilFestlegenEllipse(x: -11, y: -35, breite: 9, höhe: 9, farbe: "schwarz")
        FigurteilFestlegenEllipse(x: 2, y: -35, breite: 9, höhe: 9, farbe: "schwarz")
    }
    
    /**
	Der Chaot nimmt dem Monster ein Leben ab.
	- parameters:
		- monster: das Monster
    */
    override func Reagieren(monster: Monster)
    {
        if (Berührt(objekt: monster))
        {
            monster.LebenNehmen()
            AufStartpositionSetzen()
        }
    }
    
    /**
    Setzt den Chaot auf die Startposition zurück
    */
    private func AufStartpositionSetzen()
    {
        PositionSetzen(x: xStart * rahmen.basisGröße +  rahmen.basisGröße / 2, y: yStart * rahmen.basisGröße +  rahmen.basisGröße / 2)
    }
    
    /**
	Berechnet den Kästchenabstand der gegebenen Koordinaten zum Monster
	- parameters:
		- x: x-Koordinate des Punktes, zu dem der Abstand berechnet wird
		- y: y-Koordinate des Punktes, zu dem der Abstand berechnet wird
	- returns: Kästchenabstand
    */
    private func AbstandGeben(x: Int, y: Int) -> Int
    {
        return (abs(x - rahmen.MonsterGeben().XPositionGeben()) + abs(y - rahmen.MonsterGeben().YPositionGeben())) / rahmen.basisGröße
    }
    
    /**
    Der Chaot bewegt sich direkt auf das Monster zu,
	bei nahem Abstand zufällig
    */
    private func Bewegen()
    {
        var minAbstand = 100000
        var xOpt = XPositionGeben()
        var yOpt = YPositionGeben()
        if AbstandGeben(x: XPositionGeben(), y: YPositionGeben()) <= 10
        {
            repeat
            {
                switch zufall.ZufallGanz(top: 4)
                {
                    case 0:
                        xOpt = XPositionGeben() + rahmen.basisGröße
                        yOpt = YPositionGeben()
                    case 1:
                        xOpt = XPositionGeben() - rahmen.basisGröße
                        yOpt = YPositionGeben()
                    case 2:
                        xOpt = XPositionGeben()
                        yOpt = YPositionGeben() + rahmen.basisGröße
                    case 3:
                        xOpt = XPositionGeben()
                        yOpt = YPositionGeben() - rahmen.basisGröße
                    default:
                        break
                }
            }
            while (rahmen.IstMauer(x: xOpt, y: yOpt))
        }
        else
        {
            if !rahmen.IstMauer(x: XPositionGeben(), y: YPositionGeben() - rahmen.basisGröße)
            {
                xOpt = XPositionGeben()
                yOpt = YPositionGeben() - rahmen.basisGröße
                minAbstand = AbstandGeben(x: xOpt, y: yOpt)
            }
            if (AbstandGeben(x: XPositionGeben(), y: YPositionGeben() + rahmen.basisGröße) < minAbstand) && (!rahmen.IstMauer(x: XPositionGeben(), y: YPositionGeben() + rahmen.basisGröße))
            {
                xOpt = XPositionGeben()
                yOpt = YPositionGeben() + rahmen.basisGröße
                minAbstand = AbstandGeben(x: xOpt, y: yOpt)
            }
            if (AbstandGeben(x: XPositionGeben() + rahmen.basisGröße, y: YPositionGeben()) < minAbstand) && (!rahmen.IstMauer(x: XPositionGeben() + rahmen.basisGröße, y: YPositionGeben()))
            {
                xOpt = XPositionGeben() + rahmen.basisGröße
                yOpt = YPositionGeben()
                minAbstand = AbstandGeben(x: xOpt, y: yOpt)
            }
            if (AbstandGeben(x: XPositionGeben() - rahmen.basisGröße, y: YPositionGeben()) < minAbstand) && (!rahmen.IstMauer(x: XPositionGeben() - rahmen.basisGröße, y: YPositionGeben()))
            {
                xOpt = XPositionGeben() - rahmen.basisGröße
                yOpt = YPositionGeben()
                minAbstand = AbstandGeben(x: xOpt, y: yOpt)
            }
        }
        PositionSetzen(x: xOpt, y: yOpt)
    }
    
    /**
    Der Chaot geht
    */
    override func AktionAusführen()
    {
        Bewegen()
        if Berührt(objekt: rahmen.MonsterGeben())
        {
            rahmen.MonsterGeben().LebenNehmen()
            AufStartpositionSetzen()
        }
    }
}

/**
Geht konsequent zu einem Punkt 4 Einheiten vor dem Monster.

- author: Albert Wiedemann
- version: 1.0
*/
class Fallensteller: Hindernis
{
    /** x-Startposition merken */
    var xStart: Int
    /** y-Startposition merken */
    var yStart: Int
    
    /**
	Der Konstruktor legt die Position fest
	- parameters:
		- x: x-Koordinate des Hindernisses
		- y: y-Koordinate des Hindernisses
		- rahmen: das Rahmenobjekt des Spiels
    */
    override init(x: Int, y: Int, rahmen: Rahmen)
    {
        xStart = x
        yStart = y
        super.init(x: x, y: y, rahmen: rahmen)
        FigurErzeugen(farbe: "grün")
        GrößeSetzen(größe: rahmen.basisGröße)
    }
    
    /**
	Erzeugt das Figurenbild in der gewünschten Farbe
	- parameters:
		- farbe: die gewünschte Farbe
    */
    private func FigurErzeugen(farbe: String)
    {
        EigeneFigurLöschen()
        FigurteilFestlegenEllipse(x: -15, y: -50, breite: 30, höhe: 30, farbe: farbe)
        FigurteilFestlegenDreieck(x1: -25, y1: 45, x2: 0, y2: -50, x3: 25, y3: 45, farbe: farbe)
        FigurteilFestlegenRechteck(x: -15, y: -35, breite: 30, höhe: 80, farbe: farbe)
        FigurteilFestlegenEllipse(x: -25, y: 40, breite: 10, höhe: 10, farbe: farbe)
        FigurteilFestlegenEllipse(x: -15, y: 40, breite: 10, höhe: 10, farbe: farbe)
        FigurteilFestlegenEllipse(x: -5, y: 40, breite: 10, höhe: 10, farbe: farbe)
        FigurteilFestlegenEllipse(x: 5, y: 40, breite: 10, höhe: 10, farbe: farbe)
        FigurteilFestlegenEllipse(x: 15, y: 40, breite: 10, höhe: 10, farbe: farbe)
        FigurteilFestlegenEllipse(x: -11, y: -35, breite: 9, höhe: 9, farbe: "schwarz")
        FigurteilFestlegenEllipse(x: 2, y: -35, breite: 9, höhe: 9, farbe: "schwarz")
    }
    
    /**
	Der Fallensteller nimmt dem Monster ein Leben ab.
	- parameters:
		- monster: das Monster
    */
    override func Reagieren(monster: Monster)
    {
        if (Berührt(objekt: monster))
        {
            monster.LebenNehmen()
            AufStartpositionSetzen()
        }
    }
    
    /**
    Setzt den Fallensteller auf die Startposition zurück
    */
    private func AufStartpositionSetzen()
    {
        PositionSetzen(x: xStart * rahmen.basisGröße +  rahmen.basisGröße / 2, y: yStart * rahmen.basisGröße + rahmen.basisGröße / 2)
    }
    
    /**
	Berechnet den Kästchenabstand der gegebenen Koordinatenpaare
	- parameters:
		- x: x-Koordinate des 1. Punktes
		- y: y-Koordinate des 1. Punktes
		- x2: x-Koordinate des 2. Punktes
		- y2: y-Koordinate des 2. Punktes
	- returns: Kästchenabstand
    */
    private func AbstandGeben(x: Int, y: Int, x2: Int, y2: Int) -> Int
    {
        return (abs(x - x2) + abs(y - y2)) / rahmen.basisGröße
    }
    
    /**
    Der Fallensteller bewegt sich auf einen Punkt vor dem Monster zu
    */
    private func Bewegen()
    {
        var minAbstand = 100000
        var xZiel = rahmen.MonsterGeben().XPositionGeben()
        var yZiel = rahmen.MonsterGeben().YPositionGeben()
        switch rahmen.MonsterGeben().WinkelGeben()
        {
            case 0:
                xZiel = xZiel + 4 * rahmen.basisGröße
            case 90:
                yZiel = yZiel - 4 * rahmen.basisGröße
            case 180:
                xZiel = xZiel - 4 * rahmen.basisGröße
            case 270:
                yZiel = yZiel + 4 * rahmen.basisGröße
            default:
                break
        }
        var xOpt = XPositionGeben()
        var yOpt = YPositionGeben()
        if !rahmen.IstMauer(x: XPositionGeben(), y: YPositionGeben() - rahmen.basisGröße)
        {
            xOpt = XPositionGeben()
            yOpt = YPositionGeben() - rahmen.basisGröße
            minAbstand = AbstandGeben(x: xOpt, y: yOpt, x2: xZiel, y2: yZiel)
        }
        if (AbstandGeben(x: XPositionGeben(), y: YPositionGeben() + rahmen.basisGröße, x2: xZiel, y2: yZiel) < minAbstand) && (!rahmen.IstMauer(x: XPositionGeben(), y: YPositionGeben() + rahmen.basisGröße))
        {
            xOpt = XPositionGeben()
            yOpt = YPositionGeben() + rahmen.basisGröße
            minAbstand = AbstandGeben(x: xOpt, y: yOpt, x2: xZiel, y2: yZiel)
        }
        if (AbstandGeben(x: XPositionGeben() + rahmen.basisGröße, y: YPositionGeben(), x2: xZiel, y2: yZiel) < minAbstand) && (!rahmen.IstMauer(x: XPositionGeben() + rahmen.basisGröße, y: YPositionGeben()))
        {
            xOpt = XPositionGeben() + rahmen.basisGröße;
            yOpt = YPositionGeben();
            minAbstand = AbstandGeben(x: xOpt, y: yOpt, x2: xZiel, y2: yZiel);
        }
        if (AbstandGeben(x: XPositionGeben() - rahmen.basisGröße, y: YPositionGeben(), x2: xZiel, y2: yZiel) < minAbstand) && (!rahmen.IstMauer(x: XPositionGeben() - rahmen.basisGröße, y: YPositionGeben()))
        {
            xOpt = XPositionGeben() - rahmen.basisGröße
            yOpt = YPositionGeben()
            minAbstand = AbstandGeben(x: xOpt, y: yOpt, x2: xZiel, y2: yZiel)
        }
        PositionSetzen(x: xOpt, y: yOpt)
    }
    
    /**
    Der Fallensteller geht
    */
    override func AktionAusführen()
    {
        Bewegen()
        if Berührt(objekt: rahmen.MonsterGeben())
        {
            rahmen.MonsterGeben().LebenNehmen()
            AufStartpositionSetzen()
        }
    }
}

/**
Verfolgt das Monster konsequent

- author: Albert Wiedemann
- version: 1.0
*/
class Jäger: Hindernis
{
    /** x-Startposition merken */
    var xStart: Int
    /** y-Startposition merken */
    var yStart: Int
    
    /**
	Der Konstruktor legt die Position fest
	- parameters:
		- x: x-Koordinate des Hindernisses
		- y: y-Koordinate des Hindernisses
		- rahmen: das Rahmenobjekt des Spiels
    */
    override init(x: Int, y: Int, rahmen: Rahmen)
    {
        xStart = x
        yStart = y
        super.init(x: x, y: y, rahmen: rahmen)
        FigurErzeugen(farbe: "rot")
        GrößeSetzen(größe: rahmen.basisGröße)
    }
    
    /**
	Erzeugt das Figurenbild in der gewünschten Farbe
	- parameters:
		- farbe: die gewünschte Farbe
    */
    private func FigurErzeugen(farbe: String)
    {
        EigeneFigurLöschen()
        FigurteilFestlegenEllipse(x: -15, y: -50, breite: 30, höhe: 30, farbe: farbe)
        FigurteilFestlegenDreieck(x1: -25, y1: 45, x2: 0, y2: -50, x3: 25, y3: 45, farbe: farbe)
        FigurteilFestlegenRechteck(x: -15, y: -35, breite: 30, höhe: 80, farbe: farbe)
        FigurteilFestlegenEllipse(x: -25, y: 40, breite: 10, höhe: 10, farbe: farbe)
        FigurteilFestlegenEllipse(x: -15, y: 40, breite: 10, höhe: 10, farbe: farbe)
        FigurteilFestlegenEllipse(x: -5, y: 40, breite: 10, höhe: 10, farbe: farbe)
        FigurteilFestlegenEllipse(x: 5, y: 40, breite: 10, höhe: 10, farbe: farbe)
        FigurteilFestlegenEllipse(x: 15, y: 40, breite: 10, höhe: 10, farbe: farbe)
        FigurteilFestlegenEllipse(x: -11, y: -35, breite: 9, höhe: 9, farbe: "schwarz")
        FigurteilFestlegenEllipse(x: 2, y: -35, breite: 9, höhe: 9, farbe: "schwarz")
    }
    
    /**
	Der Jäger nimmt dem Monster ein Leben ab.
	- parameters:
		- monster: das Monster
    */
    override func Reagieren(monster: Monster)
    {
        if (Berührt(objekt: monster))
        {
            monster.LebenNehmen()
            AufStartpositionSetzen()
        }
    }
    
    /**
    Setzt den Jäger auf die Startposition zurück
    */
    private func AufStartpositionSetzen()
    {
        PositionSetzen(x: xStart * rahmen.basisGröße +  rahmen.basisGröße / 2, y: yStart * rahmen.basisGröße +  rahmen.basisGröße / 2)
    }
    
    /**
	Berechnet den Kästchenabstand der gegebenen Koordinaten zum Monster
	- parameters:
		- x: x-Koordinate des Punktes, zu dem der Abstand berechnet wird
		- y: y-Koordinate des Punktes, zu dem der Abstand berechnet wird
	- returns: Kästchenabstand
    */
    private func AbstandGeben(x: Int, y: Int) -> Int
    {
        return (abs(x - rahmen.MonsterGeben().XPositionGeben()) + abs(y - rahmen.MonsterGeben().YPositionGeben())) / rahmen.basisGröße
    }
    
    /**
    Der Jäger bewegt sich direkt auf das Monster zu
    */
    private func Bewegen()
    {
        var minAbstand = 100000
        var xOpt = XPositionGeben()
        var yOpt = YPositionGeben()
        if !rahmen.IstMauer(x: XPositionGeben(), y: YPositionGeben() - rahmen.basisGröße)
        {
            xOpt = XPositionGeben()
            yOpt = YPositionGeben() - rahmen.basisGröße
            minAbstand = AbstandGeben(x: xOpt, y: yOpt)
        }
        if (AbstandGeben(x: XPositionGeben(), y: YPositionGeben() + rahmen.basisGröße) < minAbstand) && (!rahmen.IstMauer(x: XPositionGeben(), y: YPositionGeben() + rahmen.basisGröße))
        {
            xOpt = XPositionGeben()
            yOpt = YPositionGeben() + rahmen.basisGröße
            minAbstand = AbstandGeben(x: xOpt, y: yOpt)
        }
        if (AbstandGeben(x: XPositionGeben() + rahmen.basisGröße, y: YPositionGeben()) < minAbstand) && (!rahmen.IstMauer(x: XPositionGeben() + rahmen.basisGröße, y: YPositionGeben()))
        {
            xOpt = XPositionGeben() + rahmen.basisGröße
            yOpt = YPositionGeben()
            minAbstand = AbstandGeben(x: xOpt, y: yOpt)
        }
        if (AbstandGeben(x: XPositionGeben() - rahmen.basisGröße, y: YPositionGeben()) < minAbstand) && (!rahmen.IstMauer(x: XPositionGeben() - rahmen.basisGröße, y: YPositionGeben()))
        {
            xOpt = XPositionGeben() - rahmen.basisGröße
            yOpt = YPositionGeben()
            minAbstand = AbstandGeben(x: xOpt, y: yOpt)
        }
        PositionSetzen(x: xOpt, y: yOpt)
    }
    
    /**
    Der Jäger geht
    */
    override func AktionAusführen()
    {
        Bewegen()
        if Berührt(objekt: rahmen.MonsterGeben())
        {
            rahmen.MonsterGeben().LebenNehmen()
            AufStartpositionSetzen()
        }
    }
}

/**
Schafft den Rahmen für das Spiel.

- author: Albert Wiedemann
- version: 1.0
*/
class Rahmen: Ereignisbehandlung
{
    /** Die Seitenlänge der Basisquadrates des Spielfelds */
    let basisGröße = 20
    /** Anzahl der verstreuten Krümel. */
    var krümelAnzahl: Int
    /** Anzahl der gefressenen Krümel. */
    var gefressenAnzahl: Int
    /** Feld aller Hindernisse */
    private var hindernisse: [Hindernis]
    /** Das Monster */
    private var monster: Monster!
    /** Der Statustext */
    private var status: Status
    /** Der Jäger */
    private var jäger: Jäger!
    /** Der Chaot */
    private var chaot: Chaot!
    /** Der Fallensteller */
    private var fallensteller: Fallensteller!

    /**
    Legt das Labyrinth, die Krümel und das Monster an
    */
    override init()
    {
        krümelAnzahl = 0
        gefressenAnzahl = 0
        hindernisse = [Hindernis]()
        status = Status()
        monster = nil
        jäger = nil
        chaot = nil
        fallensteller = nil
        super.init()

        for x in 0 ..< 28
        {
            hindernisse.append(MauerRechteck(x: x, y: 0, rahmen: self))
            hindernisse.append(MauerRechteck(x: x, y: 30, rahmen: self))
        }
        for y in 1 ... 8
        {
            hindernisse.append(MauerRechteck(x: 0, y: y, rahmen: self))
            hindernisse.append(MauerRechteck(x: 27, y: y, rahmen: self))
        }
        for x in 2 ... 5
        {
            for y in 2 ... 7
            {
                hindernisse.append(MauerRechteck(x: x, y: y, rahmen: self))
                hindernisse.append(MauerRechteck(x: x + 20, y: y, rahmen: self))
            }
        }
        for x in 7 ... 9
        {
            for y in 2 ... 4
            {
                hindernisse.append(MauerRechteck(x: x, y: y, rahmen: self))
                hindernisse.append(MauerRechteck(x: x + 11, y: y, rahmen: self))
            }
        }
        for x in 11 ... 16
        {
            for y in 2 ... 4
            {
                hindernisse.append(MauerRechteck(x: x, y: y, rahmen: self))
            }
        }
        for x in 0 ... 5
        {
            for y in 9 ... 13
            {
                hindernisse.append(MauerRechteck(x: x, y: y, rahmen: self))
                hindernisse.append(MauerRechteck(x: x + 22, y: y, rahmen: self))
            }
        }
        for x in 7 ... 20
        {
            hindernisse.append(MauerRechteck(x: x, y: 6, rahmen: self))
            hindernisse.append(MauerRechteck(x: x, y: 7, rahmen: self))
        }
        for x in 7 ... 8
        {
            for y in 8 ... 10
            {
                hindernisse.append(MauerRechteck(x: x, y: y, rahmen: self))
                hindernisse.append(MauerRechteck(x: x + 12, y: y, rahmen: self))
            }
        }
        for x in 10 ... 17
        {
            hindernisse.append(MauerRechteck(x: x, y: 9, rahmen: self))
            hindernisse.append(MauerRechteck(x: x, y: 10, rahmen: self))
        }
        for y in 12 ... 16
        {
            hindernisse.append(MauerRechteck(x: 7, y: y, rahmen: self))
            hindernisse.append(MauerRechteck(x: 8, y: y, rahmen: self))
            hindernisse.append(MauerRechteck(x: 19, y: y, rahmen: self))
            hindernisse.append(MauerRechteck(x: 20, y: y, rahmen: self))
        }
        for x in 10 ... 17
        {
            for y in 12 ... 16
            {
                hindernisse.append(MauerRechteck(x: x, y: y, rahmen: self))
            }
        }
        for y in 15 ... 29
        {
            hindernisse.append(MauerRechteck(x: 0, y: y, rahmen: self))
            hindernisse.append(MauerRechteck(x: 27, y: y, rahmen: self))
        }
        for x in 1 ... 5
        {
            for y in 15 ... 19
            {
                hindernisse.append(MauerRechteck(x: x, y: y, rahmen: self))
                hindernisse.append(MauerRechteck(x: x + 21, y: y, rahmen: self))
            }
        }
        for x in 7 ... 11
        {
            hindernisse.append(MauerRechteck(x: x, y: 18, rahmen: self))
            hindernisse.append(MauerRechteck(x: x, y: 19, rahmen: self))
            hindernisse.append(MauerRechteck(x: x + 9, y: 18, rahmen: self))
            hindernisse.append(MauerRechteck(x: x + 9, y: 19, rahmen: self))
        }
        for y in 18 ... 22
        {
            hindernisse.append(MauerRechteck(x: 13, y: y, rahmen: self))
            hindernisse.append(MauerRechteck(x: 14, y: y, rahmen: self))
        }
        for x in 1 ... 2
        {
            for y in 21 ... 22
            {
                hindernisse.append(MauerRechteck(x: x, y: y, rahmen: self))
                hindernisse.append(MauerRechteck(x: x + 24, y: y, rahmen: self))
            }
        }
        for x in 4 ... 5
        {
            for y in 21 ... 25
            {
                hindernisse.append(MauerRechteck(x: x, y: y, rahmen: self))
                hindernisse.append(MauerRechteck(x: x + 18, y: y, rahmen: self))
            }
        }
        for x in 2 ... 3
        {
            for y in 24 ... 28
            {
                hindernisse.append(MauerRechteck(x: x, y: y, rahmen: self))
                hindernisse.append(MauerRechteck(x: x + 22, y: y, rahmen: self))
            }
        }
        for x in 7 ... 11
        {
            hindernisse.append(MauerRechteck(x: x, y: 21, rahmen: self))
            hindernisse.append(MauerRechteck(x: x, y: 22, rahmen: self))
            hindernisse.append(MauerRechteck(x: x + 9, y: 21, rahmen: self))
            hindernisse.append(MauerRechteck(x: x + 9, y: 22, rahmen: self))
        }
        for x in 10 ... 17
        {
            hindernisse.append(MauerRechteck(x: x, y: 24, rahmen: self))
            hindernisse.append(MauerRechteck(x: x, y: 25, rahmen: self))
        }
        hindernisse.append(MauerRechteck(x: 10, y: 23, rahmen: self))
        hindernisse.append(MauerRechteck(x: 11, y: 23, rahmen: self))
        hindernisse.append(MauerRechteck(x: 16, y: 23, rahmen: self))
        hindernisse.append(MauerRechteck(x: 17, y: 23, rahmen: self))
        for x in 7 ... 8
        {
            for y in 24 ... 25
            {
                hindernisse.append(MauerRechteck(x: x, y: y, rahmen: self))
                hindernisse.append(MauerRechteck(x: x + 12, y: y, rahmen: self))
            }
        }
        for x in 5 ... 22
        {
            hindernisse.append(MauerRechteck(x: x, y: 27, rahmen: self))
            hindernisse.append(MauerRechteck(x: x, y: 28, rahmen: self))
        }
        
        monster = Monster(x: 4, y: 1, rahmen: self, leben: 3)
        
        for x in 1 ... 26
        {
            hindernisse.append(Krümel(x: x, y: 1, rahmen: self))
            krümelAnzahl = krümelAnzahl + 1
        }
        for y in 2 ... 8
        {
            hindernisse.append(Krümel(x: 1, y: y, rahmen: self))
            hindernisse.append(Krümel(x: 26, y: y, rahmen: self))
            krümelAnzahl = krümelAnzahl + 2
        }
        for x in 7 ... 20
        {
            hindernisse.append(Krümel(x: x, y: 5, rahmen: self))
            krümelAnzahl = krümelAnzahl + 1
        }
        for y in 2 ... 4
        {
            hindernisse.append(Krümel(x: 10, y: y, rahmen: self))
            hindernisse.append(Krümel(x: 17, y: y, rahmen: self))
            krümelAnzahl = krümelAnzahl + 2
        }
        for y in 2 ... 25
        {
            hindernisse.append(Krümel(x: 6, y: y, rahmen: self))
            hindernisse.append(Krümel(x: 21, y: y, rahmen: self))
            krümelAnzahl = krümelAnzahl + 2
        }
        for x in 2 ... 5
        {
            hindernisse.append(Krümel(x: x, y: 8, rahmen: self))
            hindernisse.append(Krümel(x: x + 20, y: 8, rahmen: self))
            krümelAnzahl = krümelAnzahl + 2
        }
        for x in 9 ... 18
        {
            hindernisse.append(Krümel(x: x, y: 8, rahmen: self))
            krümelAnzahl = krümelAnzahl + 1
        }
        for x in 0 ..< 5
        {
            hindernisse.append(Krümel(x: x + 1, y: 20, rahmen: self))
            hindernisse.append(Krümel(x: x + 7, y: 20, rahmen: self))
            hindernisse.append(Krümel(x: x + 16, y: 20, rahmen: self))
            hindernisse.append(Krümel(x: x + 22, y: 20, rahmen: self))
            krümelAnzahl = krümelAnzahl + 4
        }
        for y in 20 ... 23
        {
            hindernisse.append(Krümel(x: 12, y: y, rahmen: self))
            hindernisse.append(Krümel(x: 15, y: y, rahmen: self))
            krümelAnzahl = krümelAnzahl + 2
        }
        for x in 4 ... 23
        {
            hindernisse.append(Krümel(x: x, y: 26, rahmen: self))
            krümelAnzahl = krümelAnzahl + 1
        }
        for x in 1 ... 26
        {
            hindernisse.append(Krümel(x: x, y: 29, rahmen: self))
            krümelAnzahl = krümelAnzahl + 1
        }
        for y in 23 ... 28
        {
            hindernisse.append(Krümel(x: 1, y: y, rahmen: self))
            hindernisse.append(Krümel(x: 26, y: y, rahmen: self))
            krümelAnzahl = krümelAnzahl + 2
        }
        for y in 21 ... 23
        {
            hindernisse.append(Krümel(x: 3, y: y, rahmen: self))
            hindernisse.append(Krümel(x: 24, y: y, rahmen: self))
            krümelAnzahl = krümelAnzahl + 2
        }
        hindernisse.append(Krümel(x: 2, y: 23, rahmen: self))
        hindernisse.append(Krümel(x: 25, y: 23, rahmen: self))
        krümelAnzahl = krümelAnzahl + 2
        for y in 27 ... 29
        {
            hindernisse.append(Krümel(x: 4, y: y, rahmen: self))
            hindernisse.append(Krümel(x: 23, y: y, rahmen: self))
            krümelAnzahl = krümelAnzahl + 2
        }
        for y in 23 ... 25
        {
            hindernisse.append(Krümel(x: 9, y: y, rahmen: self))
            hindernisse.append(Krümel(x: 18, y: y, rahmen: self))
            krümelAnzahl = krümelAnzahl + 2
        }
        hindernisse.append(Krümel(x: 7, y: 23, rahmen: self))
        hindernisse.append(Krümel(x: 8, y: 23, rahmen: self))
        hindernisse.append(Krümel(x: 19, y: 23, rahmen: self))
        hindernisse.append(Krümel(x: 20, y: 23, rahmen: self))
        hindernisse.append(Krümel(x: 13, y: 23, rahmen: self))
        hindernisse.append(Krümel(x: 14, y: 23, rahmen: self))
        krümelAnzahl = krümelAnzahl + 6
        
        status = Status()
        status.AnzeigeSetzen(gefressen: gefressenAnzahl, gesamt: krümelAnzahl, leben: monster.LebenGeben())
        
        jäger = Jäger (x: 13, y: 11, rahmen: self)
        hindernisse.append(jäger)
        chaot = Chaot (x: 14, y: 11, rahmen: self)
        hindernisse.append(chaot)
        fallensteller = Fallensteller (x: 12, y: 11, rahmen: self)
        hindernisse.append(fallensteller)

        Starten()
    }
    
    /**
    Meldet eine Kopie des Felds der Hindernisse zurück
	- returns Feld der Hindernisse
    */
    func HindernisseGeben() -> [Hindernis]
    {
        return hindernisse
    }
    
    /**
	Entfernt den angegebenen Krümel aus dem Feld und der Anzeige
	- parameters:
		- krümel: der zu entfernende Krümel
    */
    func KrümelEntfernen(krümel: Hindernis)
    {
        gefressenAnzahl = gefressenAnzahl + 1
        hindernisse.remove(at: hindernisse.firstIndex(where: {$0 === krümel})!)
        status.AnzeigeSetzen(gefressen: gefressenAnzahl, gesamt: krümelAnzahl, leben: monster.LebenGeben())
        if gefressenAnzahl == krümelAnzahl
        {
            Gewonnen()
            Anhalten()
        }
    }
    
    /**
    Wird vom Monster aufgerufen, wenn die Anzahl der Leben verändert wurde.
    */
    func StatusÄndern()
    {
        status.AnzeigeSetzen(gefressen: gefressenAnzahl, gesamt: krümelAnzahl, leben: monster.LebenGeben())
        if monster.LebenGeben() <= 0
        {
            Verloren()
            Anhalten()
        }
    }
    
    /**
    Meldet das Monster
    - returns: Monster
    */
    func MonsterGeben() -> Monster
    {
        return monster
    }
    
    /**
	Meldet, ob auf dem angegebenen Feld ein Mauerstein steht
	- parameters:
		- x: x-Position des Feldes
		- y: y-Position des Feldes
	- returns: wahr, wenn auf dem Feld ein Mauerstein steht
    */
    func IstMauer(x: Int, y: Int) -> Bool
    {
        if (x <= 0) || (x >= 28 * basisGröße) || (y <= 0) || (y >= 31 * basisGröße)
        {
            return true
        }
        for h in hindernisse
        {
            if (h is MauerRechteck) && (x == h.XPositionGeben()) &&  (y == h.YPositionGeben())
            {
                return true
            }
        }
        return false
    }
    
    /**
    nichts, verhindert Ausgaben der Oberklassenimplementierung
    */
    override func SonderTasteGedrückt(taste: Int)
    {
    }
    
    /**
    nichts, verhindert Ausgaben der Oberklassenimplementierung
    */
    override func TasteGedrückt(taste: Character)
    {
    }
    
    /**
    nichts, verhindert Ausgaben der Oberklassenimplementierung
    */
    override func TaktImpulsAusführen()
    {
    }
}

Rahmen()

