/*:
## **Aufgabe Snake Prototyp 3**
*/

/**
Snake

- author: Albert Wiedemann
- version 1.0
 */
import Foundation
#if os(macOS)
import Cocoa
import AppKit
#elseif os(iOS)
import UIKit
#endif
import PlaygroundSupport

#if os(macOS)
typealias SystemFarbe = NSColor
public typealias View = NSView
typealias Button = NSButton
typealias Slider = NSSlider
#elseif os(iOS)
typealias SystemFarbe = UIColor
public typealias View = UIView
typealias Button = UIButton
typealias Slider = UISlider
#endif

/**
stellt einen Zufallsgenerator zur Verfügung.
*/
class Zufall
{
	/**
    Initialisiert die verwendete Instanz des System-Zufallsgenerators
    */
    init()
    {
    	srandom(0)
    }
    
    /**
    Typische Kommazufallszahl
    - returns: Eine Zufallszahl aus dem Intervall [0; 1[
    */
    func ZufallKomma() -> Float
    {
    	return Float(arc4random()) / Float(UINT32_MAX)
    }

    /**
    Typische ganze Zufallszahl
    - returns: Eine ganze Zufallszahl aus dem Intervall [0; top[
    */
    func ZufallGanz(top: Int) -> Int
    {
    	return Int(ZufallKomma() * Float(top))
    }
}

/** Zeicheneinheit */
let einheiten = (x: CGFloat(26), y: CGFloat(20))

/**
Basisklasse für alle in der Hauptview darzustellenden Graphikobjekte.
Die Koordinaten beziehen sich auf den Mittelpunkt des Rahmenrechtecks der View.
*/
class GraphikSymbol
{
    /* Für die Schlange benötigte Farben */
    let gelb = SystemFarbe.yellow.cgColor
    let grün = SystemFarbe(red: 0.2, green: 0.78, blue: 0, alpha: 1).cgColor
    let dunkelrot = SystemFarbe(red: 0.78, green: 0, blue: 0, alpha: 1).cgColor
    let rot = SystemFarbe(red: 1.0, green: 0, blue: 0, alpha: 1).cgColor
    let grau = SystemFarbe(red: 0.78, green: 0.78, blue: 0.78, alpha: 1).cgColor
    let weiß = SystemFarbe.white.cgColor
    let blau = SystemFarbe.blue.cgColor
    let braun = SystemFarbe(red: 0.78, green: 0.39, blue: 0.0, alpha: 1).cgColor
    /* x- und y-Koordinate der linken, oberen Ecke der Figur als Rasterkoordinaten*/
    var x, y: CGFloat

    /**
    Standardkonstruktor für Objekte der Klasse GraphikSymbol.
    Farbe und Position werden vorbesetzt.
    Die View wird gegebenenfalls angelegt.
    */
	init()
    {
        x = 0
        y = 0
        if Oberfläche.g == nil
        {
        	PlaygroundPage.current.liveView = Oberfläche(frame: CGRect(x: 0, y: 0, width: Oberfläche.breiteFenster, height: Oberfläche.höheFenster + Oberfläche.höheBedienleiste))
        }
        HauptView.v.figuren.append(self)
        HauptView.v.NeuZeichnen()
    }

    /**
    Entfernt das Symbol aus der Anzeige
    */
    func Entfernen()
    {
        if let index = HauptView.v.figuren.firstIndex(where: {$0 === self})
        {
        	HauptView.v.figuren.remove(at: index)
            HauptView.v.NeuZeichnen()
        }
    }
    
    /**
    Setzt die Koordinaten der Figur auf die angegebenen Koordinaten
    - parameters:
    	- x: x-Koordinate
    	- y: y-Koordinate
    */
    func PositionSetzen(x: Int, y: Int)
    {
    	self.x = CGFloat(x)
        self.y = CGFloat(y)
        HauptView.v?.NeuZeichnen()
    }

	/**
    Meldet die x-Koordinate der Figur
    - returns: x-Koordinate
    */
    func XPositionGeben() -> Int
    {
    	return Int(x)
    }
    
	/**
    Meldet die y-Koordinate der Figur
    - returns: y-Koordinate
    */
    func YPositionGeben() -> Int
    {
    	return Int(y)
    }

    /**
    Zeichnet das Symbol in den gegebenen Context.
    Muss in den Unterklassen überschrieben werden.
    - parameters:
    	- context: Der Grafikkontext
    	- breite: die Breite der View
    	- höhe: die Höhe der View
    	- größe: die Größe der Figur (Rastergröße)
    */
	func Zeichnen(context: CGContext, breite: CGFloat, höhe: CGFloat, größe: CGFloat)
    {
    	fatalError("Zeichnen nicht überschreiben")
    }
}

/**
Methoden für den Aufruf beim Touch
*/
protocol KlickInterface
{
	/**
	Interpretiert den Mausklick / Touch als Bewegunsrichtung
	- parameters:
		- richtung: Die Himmelsrichtung
     */
	func RichtungSetzen(richtung: Character)
	/**
	Interpretiert den Mausklick / Touch als Drehanweisung
	- parameters:
		- rechts: bei wahr Rechtsdrehuhng, sonst Linksdrehung
     */
    func DrehungSetzen(rechts: Bool)
}

/**
Die Rahmenview für die Darstellung der Graphikobjekte.

Das Koordinatensystem der View ist ein Rastersystem mit Ursprung in der Mitte,
(0/0) bezeichnet das Feld rechts unterhalb des Ursprungs.
Die y-Achse zeigt nach unten
*/
class HauptView: View
{
    #if os(macOS)
    /** Hintergrundfarbe */
    var backgroundColor: SystemFarbe
    override var acceptsFirstResponder: Bool {get {return true}}
    #endif
    /** Die Hauptview für die Darstellung des Spielfeldes. */
    static var v: HauptView! = nil
    /** alle aktuell vorhandenen Figuren */
    var figuren: [GraphikSymbol] = [GraphikSymbol]()
	/** Dahin gehen die Oberflächenereignisse */
	private var klickInterface: KlickInterface! = nil

	/**
    Der Konstruktor belegt die Variable für die Anzeige und setzt die Hintergrundfarbe.
    - parameters:
    	- frame: der Platz für die View
    */
	override init(frame: CGRect)
    {
        #if os(macOS)
        backgroundColor = SystemFarbe(red: 0.98, green: 0.98, blue: 0.98, alpha: 1.0)
        #endif
        super.init(frame: frame)
        #if os(iOS)
        backgroundColor = SystemFarbe(red: 0.98, green: 0.98, blue: 0.98, alpha: 1.0)
        #endif
        HauptView.v = self
    }

	/**
    Der alternative Konstruktor belegt die Variable für die Anzeige und setzt die Hintergrundfarbe.
    - parameters:
    	- coder: Das zur Verfügung gestellte Objekt der Klasse NSCoder
    */
	required init?(coder: NSCoder)
    {
        #if os(macOS)
        backgroundColor = SystemFarbe(red: 0.98, green: 0.98, blue: 0.98, alpha: 1.0)
        #endif
    	super.init(coder: coder)
        #if os(iOS)
        backgroundColor = SystemFarbe(red: 0.98, green: 0.98, blue: 0.98, alpha: 1.0)
        #endif
        HauptView.v = self
    }

	/**
	Löst das Neuzeichnen der View aus
     */
    func NeuZeichnen()
    {
        #if os(macOS)
        needsDisplay = true
        #elseif os(iOS)
        setNeedsDisplay()
        #endif
    }
    
    /**
    Die Zeichenmethode wird immer dann aufgrufen, wenn die View oder ein Teil davon neu gezeichnet werden muss.
    - parameters:
    	- rect: das "schmutzige" Rechteck
    */
    override func draw(_ rect: CGRect)
    {
        #if os(macOS)
        let context = NSGraphicsContext.current!.cgContext
        context.scaleBy(x:1, y:-1)
        context.translateBy(x:0, y:-self.bounds.height)
        context.setFillColor(backgroundColor.cgColor)
        context.fill(bounds)
        #elseif os(iOS)
    	let context = UIGraphicsGetCurrentContext()!
        #endif
        let höheFenster = self.bounds.height
        let breiteFenster = self.bounds.width
        let größe = GrößeGeben()
        let xmin = breiteFenster / 2 - einheiten.x / 2 * größe
        let xmax = breiteFenster / 2 + einheiten.x / 2 * größe
        let ymin = höheFenster / 2 - einheiten.y / 2 * größe
        let ymax = höheFenster / 2 + einheiten.y / 2 * größe
        
        context.setStrokeColor(SystemFarbe.lightGray.cgColor)
        context.beginPath()
        for i in stride(from: CGFloat(1), through: einheiten.y / 2, by: 1)
        {
            context.move(to: CGPoint(x: xmin, y: höheFenster / 2 + i * größe))
            context.addLine(to: CGPoint(x: xmax, y: höheFenster / 2 + i * größe))
            context.move(to: CGPoint(x: xmin, y: höheFenster / 2 - i * größe))
            context.addLine(to: CGPoint(x: xmax, y: höheFenster / 2 - i * größe))
        }
        for i in stride(from: CGFloat(1), through: einheiten.x / 2, by: 1)
        {
            context.move(to: CGPoint(x: breiteFenster / 2 + i * größe, y: ymin))
            context.addLine(to: CGPoint(x: breiteFenster / 2 + i * größe, y: ymax))
            context.move(to: CGPoint(x: breiteFenster / 2 - i * größe, y: ymin))
            context.addLine(to: CGPoint(x: breiteFenster / 2 - i * größe, y: ymax))
        }
        context.strokePath()
        context.setStrokeColor(SystemFarbe.black.cgColor)
        context.beginPath()
        context.move(to: CGPoint(x: xmin, y: höheFenster / 2))
        context.addLine(to: CGPoint(x: xmax, y: höheFenster / 2))
        context.move(to: CGPoint(x:breiteFenster / 2, y: ymin))
        context.addLine(to: CGPoint(x:breiteFenster / 2, y: ymax))
        context.strokePath()
        
        for figur in figuren
        {
        	figur.Zeichnen(context: context, breite: breiteFenster, höhe: höheFenster, größe: größe)
        }
    }
    
    /**
	Gibt die tatsächliche Größe einer Rastereinheit zurück.
    - returns: Rastergröße in Punkten
    */
    func GrößeGeben() -> CGFloat
    {
        let größeh = self.bounds.height / einheiten.y
        let größeb = self.bounds.width / einheiten.x
        return größeb <= größeh ? größeb : größeh
    }
    
    /**
    Setzt das Touchnterface für die View
    - parameters:
    	- interface: das zu bedienende Interface
    */
    func EreignisBehandlungSetzen(interface: KlickInterface)
    {
    	klickInterface = interface
    }

	/**
	Ruft die den Klick enstprechednen ereignisse auf.
	- parameters:
		- wo: Punkt relativ zur View
    */
	private func KlickAuswerten(wo: CGPoint)
	{
        if klickInterface != nil
        {
            klickInterface.DrehungSetzen(rechts: wo.x >= self.bounds.width / 2)
            if wo.y == 0
            {
                klickInterface.RichtungSetzen(richtung: "N")
            }
            else if wo.y == self.bounds.height
            {
                klickInterface.RichtungSetzen(richtung: "S")
            }
            else if wo.x / wo.y >= self.bounds.width / self.bounds.height
            {
                if (self.bounds.height - wo.y) / wo.x >= self.bounds.height / self.bounds.width
                {
                    klickInterface.RichtungSetzen(richtung: "N")
                }
                else
                {
                    klickInterface.RichtungSetzen(richtung: "O")
                }
            }
            else
            {
                if (self.bounds.height - wo.y) / wo.x >= self.bounds.height / self.bounds.width
                {
                    klickInterface.RichtungSetzen(richtung: "W")
                }
                else
                {
                    klickInterface.RichtungSetzen(richtung: "S")
                }
            }
        }
    }
    
    #if os(macOS)
	/**
    Wird zu Beginn eines Klicks aufgerufen.
    - parameters:
    	- touches: Menge der Berührungen
    	- event: das auslösende Ereignis
    */
    override func mouseDown(with event: NSEvent)
    {
        var wo = convert(event.locationInWindow, from: nil)
        wo.y = self.bounds.height - wo.y
        KlickAuswerten(wo: wo)
    }
    #elseif os(iOS)
    override func tintColorDidChange()
    {
		setNeedsDisplay()
	}

	/**
    Wird zu Beginn einer Serie von Berührungen aufgerufen.
    - parameters:
    	- touches: Menge der Berührungen
    	- event: das auslösende Ereignis
    */
    override func touchesBegan(_ touches: Set<UITouch>, with event: UIEvent?)
    {
        for touch in touches as Set<UITouch>
        {
        	let wo = touch.location(in: self)
        	KlickAuswerten(wo: wo)
        }
        super.touchesBegan(touches, with: event)
    }
    #endif
}

/**
Beschreibt die Schnittstelle der Oberfläche zum Kontrolleur

- author: Albert Wiedemann
- version: 1.0
*/
protocol EreignisInterface: KlickInterface
{
    /**
    Startet das Spiel
    */
    func StartAusführen()
    
    /**
    Setzt den Starttakt für das Spiel.
	- parameters:
        - starttakt: Der Starttakt in Millisekunden
    */
    func StartTaktSetzen(starttakt: Int)
}


/**
Gesamtfenster (Hauptview + Tasten/Regler)
*/
class Oberfläche: View
{
    /** Anzeigegröße für eine Rasterkachel */
    private static let rasterGröße = 30
    /** Höhe des Anzeigefensters */
    static let höheFenster = rasterGröße * Int(einheiten.y)
    /** Breite des Anzeigefensters */
    static let breiteFenster = rasterGröße * Int(einheiten.x)
    /** Zusatzhöhe für Bedienelemente */
    static let höheBedienleiste = 30

    /** die eigentliche Anzeigeview */
    private var anzeige: HauptView! = nil
    /** Referenz auf den Start-Knopf */
    private var startKnopf: Button! = nil
    /** Slider für die Taktgeschwindigkeit */
    private var sliderTakt: Slider! = nil

    /** Referenz auf das Gesamtfenster */
    static var g: Oberfläche? = nil
    /** Kontrolleur */
    private var ereignisbehandlung: EreignisInterface? = nil
    
    /*
    Standardkonstruktor
    */
    override init(frame: CGRect)
    {
        super.init(frame: frame)
        ElementeAnlegen()
        Oberfläche.g = self
    }

    /*
    Konstruktor für Erstellung aus NIB-File
    */
    required init?(coder: NSCoder)
    {
        super.init(coder: coder)
        ElementeAnlegen()
        Oberfläche.g = self
    }
    
    /**
    Kontrolleur setzen
    - parameters:
    	- e: Ereignisbehandlung
 	*/
    class func EreignisbehandlungSetzen(e: EreignisInterface)
	{
        if Oberfläche.g == nil
        {
        	PlaygroundPage.current.liveView = Oberfläche(frame: CGRect(x: 0, y: 0, width: Oberfläche.breiteFenster, height: Oberfläche.höheFenster + Oberfläche.höheBedienleiste))
        }
        Oberfläche.g!.ereignisbehandlung = e
        HauptView.v.EreignisBehandlungSetzen(interface: e)
    }
    
    /**
    Aktionsfunktion für den Start-Knopf
    - parameters:
        - g: das implementierende Gesamtfenster
    */
    @objc func TimerStart(g: Oberfläche)
    {
        ereignisbehandlung?.StartAusführen()
    }
    
    /**
    Aktionsfunktion für die Taktzeiteinstellung
    - parameters:
        - g: das implementierende Gesamtfenster
    */
    @objc func TimerWert(g: Oberfläche)
    {
		#if os(macOS)
        ereignisbehandlung?.StartTaktSetzen(starttakt: sliderTakt.integerValue)
		#elseif os(iOS)
        ereignisbehandlung?.StartTaktSetzen(starttakt: Int(sliderTakt!.value))
		#endif
    }

    /**
    Legt die Oberflächenelemente an
    */
    private func ElementeAnlegen()
    {
		#if os(macOS)
        anzeige = HauptView(frame: CGRect(x: 0, y: Oberfläche.höheBedienleiste, width: Oberfläche.breiteFenster, height: Oberfläche.höheFenster))
        addSubview(anzeige)
        HauptView.v = anzeige
        startKnopf = NSButton(title: "Start", target: self, action: #selector(Oberfläche.TimerStart))
        startKnopf.setFrameOrigin(NSPoint(x: 0, y: 0))
        startKnopf.setFrameSize(NSSize(width: 80, height: 30))
        addSubview(startKnopf)
        sliderTakt = NSSlider(value: 500, minValue: 0, maxValue: 1000, target: self, action: #selector(Oberfläche.TimerWert))
        sliderTakt.numberOfTickMarks = 10
        sliderTakt.setFrameOrigin(NSPoint(x: 180, y: 0))
        sliderTakt.setFrameSize(NSSize(width: 200, height: 30))
        addSubview(sliderTakt)
		#elseif os(iOS)
        anzeige = HauptView(frame: CGRect(x: 0, y: 0, width: Oberfläche.breiteFenster, height: Oberfläche.höheFenster))
        addSubview(anzeige)
        HauptView.v = anzeige
        startKnopf = UIButton(type: .roundedRect)
        startKnopf.addTarget(self, action: #selector(Oberfläche.TimerStart), for: .touchUpInside)
        startKnopf.frame = CGRect(origin: CGPoint(x: 0, y: Oberfläche.höheFenster), size: CGSize(width: 80, height: 30))
        startKnopf.setTitle("Start", for: .normal)
        startKnopf.isEnabled = true
        addSubview(startKnopf)
        sliderTakt = UISlider(frame: CGRect(origin: CGPoint(x: 180, y: Oberfläche.höheFenster), size: CGSize(width: 200, height: 30)))
        sliderTakt.minimumValue = 0
        sliderTakt.maximumValue = 1000
        sliderTakt.value = 500
        sliderTakt.addTarget(self, action: #selector(Oberfläche.TimerWert), for: .valueChanged)
        sliderTakt.isContinuous = false
        addSubview(sliderTakt)
		#endif
    }
}

/**
Zeichnet die Begrenzung und verwaltet die Grenzkoordinaten
*/
class RandSymbol: GraphikSymbol
{
	private var xMin = 0, xMax = 0, yMin = 0, yMax: Int = 0
    private var punkte = 0
    private var wertGesetzt: Bool
    private var endemeldungGesetzt: Bool
    private var endeMeldung: String = ""
    private var sonderAnzeigeGesetzt: Bool
    private var sonderAnzeige: String = ""

    /**
    Standardkonstruktor für Objekte der Klasse RANDSYMBOL.
    Er erzeugt ein Randsymbol in der maxialen Größe.
    */
    override init()
    {
        wertGesetzt = false
        endemeldungGesetzt = false
        sonderAnzeigeGesetzt = false
        super.init()
        GrößeSetzen(breite: Int(einheiten.x), höhe: Int(einheiten.y))
    }

    /**
    Standardkonstruktor für Objekte der Klasse RANDSYMBOL.
    Er erzeugt ein Randsymbol in der angegebenen Größe.
    - parameters:
    	- breite: Breite des Randsymbols
    	- höhe: Höhe des Randsymbols
    */
    init(breite: Int, höhe: Int)
    {
        wertGesetzt = false
        endemeldungGesetzt = false
        sonderAnzeigeGesetzt = false
        super.init()
        GrößeSetzen(breite: breite, höhe: höhe)
    }

    /**
    Setzt die Breite und Hoehe des Randelements.
    Die Maximalkoordinaten des Spielfelds werden ermittelt.
    - parameters:
    	- breite: Spielfeldbreite in Kästchen
    	- höhe: Spielfeldhöhe in Kästchen
    */
    private func GrößeSetzen(breite: Int, höhe: Int)
    {
        xMax = breite / 2 - 1
        yMax = höhe / 2 - 1
        if breite % 2 == 0
        {
            xMin = -xMax
        }
        else
        {
            xMin = -xMax - 1
        }
        if höhe % 2 == 0
        {
            yMin = -yMax
        }
        else
        {
            yMin = -yMax - 1
        }
        HauptView.v?.NeuZeichnen()
    }
    
    /**
    Gibt den minimal erlaubten x-Wert zurück.
    - returns: minimaler x-Wert
    */
    func XMinGeben() -> Int
    {
    	return xMin
    }
    
    /**
    Gibt den maximal erlaubten x-Wert zurück.
    - returns: maximaler x-Wert
    */
    func XMaxGeben() -> Int
    {
    	return xMax - 1
    }
    
    /**
    Gibt den minimal erlaubten y-Wert zurück.
    - returns: minimaler y-Wert
    */
    func YMinGeben() -> Int
    {
    	return yMin
    }
    
    /**
    Gibt den maximal erlaubten y-Wert zurück.
    - returns: maximaler y-Wert
    */
    func YMaxGeben() -> Int
    {
    	return yMax - 1
    }
    
    /**
    Setzt den Punktestand für die Anzeige
    - parameters:
    	- punkte: der anzuzeigende Punktestand
    */
    func PunkteSetzen(punkte: Int)
    {
    	self.punkte = punkte
        wertGesetzt = true
        HauptView.v?.NeuZeichnen()
    }
    
    /**
    Setzt die Endemeldung für die Anzeige
    - parameters:
    	- meldung: die anzuzeigende Endemeldung
    */
    func EndemeldungSetzen(meldung: String)
    {
    	endeMeldung = meldung
        endemeldungGesetzt = true
        HauptView.v?.NeuZeichnen()
    }
    
    /**
    Setzt die Sonderanzeige für die Anzeige
    - parameters:
    	- meldung: die anzuzeigende Sonderanzeige
    */
    func SonderanzeigeSetzen(meldung: String)
    {
    	sonderAnzeige = meldung
        sonderAnzeigeGesetzt = meldung != ""
        HauptView.v?.NeuZeichnen()
    }
    
    /**
    Zeichnet das Randsymbol in den gegebenen Context.
    - parameters:
    	- context: Der Grafikkontext
   		- breite: die Breite der View
    	- höhe: die Höhe der View
    	- größe: die Größe der Figur (Rastergröße)
    */
	override func Zeichnen(context: CGContext, breite: CGFloat, höhe: CGFloat, größe: CGFloat)
    {
        context.setFillColor(grau)
        context.move(to: CGPoint(x: 0, y: 0))
        context.addLine(to: CGPoint(x: breite, y: 0))
        context.addLine(to: CGPoint(x: breite, y: höhe))
        context.addLine(to: CGPoint(x: 0, y: höhe))
        context.addLine(to: CGPoint(x: 0, y: 0))
        context.addLine(to: CGPoint(x: breite / 2 + CGFloat(xMin) * größe, y: höhe / 2 + CGFloat(yMin) * größe))
        context.addLine(to: CGPoint(x: breite / 2 + CGFloat(xMin) * größe, y: höhe / 2 + CGFloat(yMax) * größe))
        context.addLine(to: CGPoint(x: breite / 2 + CGFloat(xMax) * größe, y: höhe / 2 + CGFloat(yMax) * größe))
        context.addLine(to: CGPoint(x: breite / 2 + CGFloat(xMax) * größe, y: höhe / 2 + CGFloat(yMin) * größe))
        context.addLine(to: CGPoint(x: breite / 2 + CGFloat(xMin) * größe, y: höhe / 2 + CGFloat(yMin) * größe))
        context.addLine(to: CGPoint(x: 0, y: 0))
        context.fillPath()

		if wertGesetzt
        {
            let font = CTFontCreateWithName("Helvetica" as CFString, 14.0, nil)
            let attrString = NSAttributedString(string: "Punkte: \(punkte)", attributes: [NSAttributedString.Key.foregroundColor:SystemFarbe.blue, NSAttributedString.Key.font:font])
            let line = CTLineCreateWithAttributedString(attrString)
            context.textMatrix  = CGAffineTransform(a: 1, b: 0.0, c: 0.0, d: -1, tx: 0.0, ty: 0.0)
            context.textPosition = CGPoint(x: 4, y: 3 * größe / 4)
            CTLineDraw(line, context)
        }
		if sonderAnzeigeGesetzt
        {
            let font = CTFontCreateWithName("Helvetica" as CFString, 14.0, nil)
            let attrString = NSAttributedString(string: sonderAnzeige, attributes: [NSAttributedString.Key.foregroundColor:SystemFarbe.green, NSAttributedString.Key.font:font])
            let line = CTLineCreateWithAttributedString(attrString)
            context.textMatrix  = CGAffineTransform(a: 1, b: 0.0, c: 0.0, d: -1, tx: 0.0, ty: 0.0)
            context.textPosition = CGPoint(x: breite - 50, y: 3 * größe / 4)
            CTLineDraw(line, context)
        }
		if endemeldungGesetzt
        {
            let font = CTFontCreateWithName("Helvetica" as CFString, 20.0, nil)
            let attrString = NSAttributedString(string: endeMeldung, attributes: [NSAttributedString.Key.foregroundColor:SystemFarbe.red, NSAttributedString.Key.font:font])
            let line = CTLineCreateWithAttributedString(attrString)
            context.textMatrix  = CGAffineTransform(a: 1, b: 0.0, c: 0.0, d: -1, tx: 0.0, ty: 0.0)
            context.textPosition = CGPoint(x: breite / 2 + CGFloat(xMin + 1) * größe, y: höhe / 4)
            CTLineDraw(line, context)
        }
    }
}

/**
Stellt das Kopfsymbol er Schlange in der Hauptview dar.
*/
class KopfSymbol: GraphikSymbol
{
    /** Ausrichtung des Kopfes */
    private var ausrichtung: Character

    /**
    Standardkonstruktor für Objekte der Klasse KOPFSYMBOL.
    Er erzeugt ein Kopfsymbol in der Zelle (0 / 0) des Fensters.
    */
	override init()
    {
        ausrichtung = "O"
        super.init()
    }
    
    /**
    Zeichnet das Kopfsymbol in den gegebenen Context.
    - parameters:
    	- context: Der Grafikkontext
   		- breite: die Breite der View
    	- höhe: die Höhe der View
    	- größe: die Größe der Figur (Rastergröße)
    */
	override func Zeichnen(context: CGContext, breite: CGFloat, höhe: CGFloat, größe: CGFloat)
    {
        let x2 = breite / 2 + x * größe
        let y2 = höhe / 2 + y * größe

        context.saveGState()
        switch ausrichtung
        {
        	case "O", "o":
	        context.concatenate(CGAffineTransform(a: 1, b: 0, c: 0, d: 1, tx: x2, ty: y2))
        	case "N", "n":
	        context.concatenate(CGAffineTransform(a: 0, b: -1, c: 1, d: 0, tx: x2, ty: y2 + größe))
        	case "W", "w":
	        context.concatenate(CGAffineTransform(a: -1, b: 0, c: 0, d: -1, tx: x2 + größe, ty: y2 + größe))
        	case "S", "s":
	        context.concatenate(CGAffineTransform(a: 0, b: 1, c: -1, d: 0, tx: x2 + größe, ty: y2))
            default:
            context.concatenate(CGAffineTransform(a: 1, b: 0, c: 0, d: 1, tx: x2, ty: y2))
        }
        context.setFillColor(gelb)
        context.beginPath()
        context.move(to: CGPoint(x: 0, y: 0))
        context.addLine(to: CGPoint(x: größe / 2, y: 0))
        context.addArc(center: CGPoint(x: größe / 2, y: größe / 2), radius: größe / 2, startAngle: -(3.1415927 / 2), endAngle: 3.1415927 / 2, clockwise: false)
        context.addLine(to: CGPoint(x: 0, y: größe))
        context.closePath()
        let p = context.path!
        context.fillPath()
        context.addPath(p)
        context.setStrokeColor(grün)
		context.strokePath()
        context.setFillColor(SystemFarbe.black.cgColor)
        context.fillEllipse(in: CGRect(x: 2 * größe / 5, y: größe / 5, width: größe / 5, height: größe / 5))
        context.fillEllipse(in: CGRect(x: 2 * größe / 5, y: 3 * größe / 5, width: größe / 5, height: größe / 5))
        context.setStrokeColor(dunkelrot)
        context.move(to: CGPoint(x: größe, y: größe / 2))
        context.addLine(to: CGPoint(x: größe * 5 / 4, y: größe / 2))
        context.addLine(to: CGPoint(x: größe * 6 / 4, y: größe / 4))
        context.move(to: CGPoint(x: größe * 5 / 4, y: größe / 2))
        context.addLine(to: CGPoint(x: größe * 6 / 4, y: größe * 3 / 4))
		context.strokePath()
        context.restoreGState()
    }
    
    /**
    Setzt die Ausrichtung des Schlangenkopfs
    - parameters:
    	- a: die (neue) Ausrichtung
    */
    func AusrichtungSetzen(a: Character)
    {
    	ausrichtung = a
    	HauptView.v?.NeuZeichnen()
    }
    
    /**
    Meldet die Ausrichtung des Schalngenkopfs
    - returns: die aktuelle Ausrichtung
    */
    func AusrichtungGeben() -> Character
    {
    	return ausrichtung
    }
}

/**
Stellt das Rumpfsymbol einer Schlange in der Hauptview dar.
*/
class RumpfSymbol: GraphikSymbol
{
    /**
    Standardkonstruktor für Objekte der Klasse RUMPFSYMBOL.
    Er erzeugt ein Rumpfsymbol in der Zelle (0 / 0) des Fensters.
    */
	override init()
    {
        super.init()
    }
    
    /**
    Zeichnet das Rumpfsymbol in den gegebenen Context.
    - parameters:
    	- context: Der Grafikkontext
   		- breite: die Breite der View
    	- höhe: die Höhe der View
    	- größe: die Größe der Figur (Rastergröße)
    */
	override func Zeichnen(context: CGContext, breite: CGFloat, höhe: CGFloat, größe: CGFloat)
    {
        let x2 = breite / 2 + x * größe
        let y2 = höhe / 2 + y * größe

        context.setFillColor(gelb)
        context.fill(CGRect(x: x2, y: y2, width: größe, height: größe))
        context.setStrokeColor(grün)
        context.setLineWidth(3)
        context.stroke(CGRect(x: x2, y: y2, width: größe, height: größe))
        context.move(to: CGPoint(x: x2, y: y2))
        context.addLine(to: CGPoint(x: x2 + größe, y: y2 + größe))
        context.move(to: CGPoint(x: x2, y: y2 + größe))
        context.addLine(to: CGPoint(x: x2 + größe, y: y2))
        context.strokePath()
    }
}

/**
Stellt das Apfelsymbol in der Hauptview dar.
*/
class ApfelSymbol: GraphikSymbol
{
    
    /**
    Standardkonstruktor für Objekte der Klasse ApfelSymbol.
    Er erzeugt ein ApfelSymbol in der Zelle (0 / 0) des Fensters.
    */
	override init()
    {
        super.init()
    }
    
    /**
    Zeichnet das ApfelSymbol in den gegebenen Context.
    - parameters:
    	- context: Der Grafikkontext
   		- breite: die Breite der View
    	- höhe: die Höhe der View
    	- größe: die Größe der Figur (Rastergröße)
    */
    override func Zeichnen(context: CGContext, breite: CGFloat, höhe: CGFloat, größe: CGFloat)
    {
        let x2 = breite / 2 + x * größe
        let y2 = höhe / 2 + y * größe

        context.setFillColor(weiß)
        context.fill(CGRect(x: x2 + 1, y: y2 + 1, width: größe - 2, height: größe - 2))
        context.setFillColor(rot)
        context.fillEllipse(in: CGRect(x: x2 + größe / 4, y: y2 + 3 * größe / 8, width: größe / 2, height: größe / 2))
        context.setStrokeColor(grün)
        context.setLineWidth(3)
        context.move(to: CGPoint(x: x2 + größe / 2, y: y2 + 3 * größe / 8))
        context.addLine(to: CGPoint(x: x2 + 5 * größe / 8, y: y2 + größe / 8))
        context.move(to: CGPoint(x: x2 + größe / 2, y: y2 + 3 * größe / 8))
        context.addLine(to: CGPoint(x: x2 + größe / 4, y: y2 + größe / 8))
        context.strokePath()
        context.setStrokeColor(braun)
        context.move(to: CGPoint(x: x2 + größe / 2, y: y2 + 3 * größe / 8))
        context.addLine(to: CGPoint(x: x2 + 3 * größe / 8, y: y2 + größe / 8))
        context.strokePath()
    }
}

/**
Stellt das Pilzsymbol in der Hauptview dar.
*/
class PilzSymbol: GraphikSymbol
{
    
    /**
    Standardkonstruktor für Objekte der Klasse PilzSymbol.
    Er erzeugt ein PilzSymbol in der Zelle (0 / 0) des Fensters.
    */
	override init()
    {
        super.init()
    }
    
    /**
    Zeichnet das Pilzsymbol in den gegebenen Context.
    - parameters:
    	- context: Der Grafikkontext
   		- breite: die Breite der View
    	- höhe: die Höhe der View
    	- größe: die Größe der Figur (Rastergröße)
    */
    override func Zeichnen(context: CGContext, breite: CGFloat, höhe: CGFloat, größe: CGFloat)
    {
        let x2 = breite / 2 + x * größe
        let y2 = höhe / 2 + y * größe

        context.setFillColor(weiß)
        context.fill(CGRect(x: x2 + 1, y: y2 + 1, width: größe - 2, height: größe - 2))
        context.setFillColor(dunkelrot)
        context.beginPath()
        context.addArc(center: CGPoint(x: x2 + größe / 2, y: y2 + größe / 2), radius: 3 * größe / 8, startAngle: 0, endAngle: 3.1415927, clockwise: true)
        context.fillPath()
        context.setFillColor(braun)
        context.fill(CGRect(x: x2 + 7 * größe / 16, y: y2 + größe / 2, width: größe / 8, height: größe / 2 - 1))
        context.setFillColor (weiß)
        context.fillEllipse(in: CGRect(x: x2 + 5 * größe / 16, y: y2 + 5 * größe / 16, width: größe / 8, height: größe / 8))
        context.fillEllipse(in: CGRect(x: x2 + 9 * größe / 16, y: y2 + 5 * größe / 16, width: größe / 8, height: größe / 8))
        context.fillEllipse(in: CGRect(x: x2 + 7 * größe / 16, y: y2 + 3 * größe / 16, width: größe / 8, height: größe / 8))
    }
}

/**
Stellt das Minussymbol in der Hauptview dar.
*/
class MinusSymbol: GraphikSymbol
{
    
    /**
    Standardkonstruktor für Objekte der Klasse Minussymbol.
    Er erzeugt ein Minussymbol in der Zelle (0 / 0) des Fensters.
    */
	override init()
    {
        super.init()
    }
    
    /**
    Zeichnet das Minussymbol in den gegebenen Context.
    - parameters:
    	- context: Der Grafikkontext
   		- breite: die Breite der View
    	- höhe: die Höhe der View
    	- größe: die Größe der Figur (Rastergröße)
    */
    override func Zeichnen(context: CGContext, breite: CGFloat, höhe: CGFloat, größe: CGFloat)
    {
        let x2 = breite / 2 + x * größe
        let y2 = höhe / 2 + y * größe

        context.setFillColor(weiß)
        context.fill(CGRect(x: x2 + 1, y: y2 + 1, width: größe - 2, height: größe - 2))
        context.setFillColor(gelb)
        context.fillEllipse (in: CGRect(x: x2 + größe / 4, y: y2 + größe / 4, width: größe / 2, height: größe / 2))
        context.setStrokeColor(blau)
        context.setLineWidth(3)
        context.move(to: CGPoint(x: x2 + größe / 4, y: y2 + größe / 2))
        context.addLine(to: CGPoint(x: x2 + 3 * größe / 4, y: y2 + größe / 2))
        context.strokePath()
    }
}

/**
Verwaltet die Ereignisse

Erzeugt Takt- und Touch-Events
*/
class EreignisBehandlung: EreignisInterface
{
	/** Referenz auf das Timer-Objekt */
	private var timer: Timer!
    /** Aktuell eingestelltes Zeitintervall in Sekunden */
    private var interval: TimeInterval

	/**
    Der Standardkonstruktor bereitet die Hauptview vor und setzt den Taktgeber.
    */
	init()
    {
    	timer = nil
        interval = 1.0
        Oberfläche.EreignisbehandlungSetzen(e: self)
    }
    /**
    Startet das Spiel
    */
    func StartAusführen()
    {
    	timer = Timer.scheduledTimer(timeInterval: interval, target: self, selector: #selector(EreignisBehandlung.Impuls), userInfo: nil, repeats: true)
    }
    
    /**
    Setzt den Starttakt für das Spiel.
	- parameters:
        - starttakt: Der Starttakt in Millisekunden
    */
    func StartTaktSetzen(starttakt: Int)
    {
    	if timer != nil
        {
            Anhalten()
            interval = Double(starttakt) / 1000.0
            StartAusführen()
        }
        else
        {
            interval = Double(starttakt) / 1000.0
        }
    }
    
    /**
    	Hält den Taktgeber an
    */
    func Anhalten()
    {
        timer.invalidate()
    	timer = nil
    }

    /**
    Interne Funktion, um den Parameter von TaktImpulsAusfuehren fernzuhalten
    - parameters:
    	- t: der aufrufende Timer
    */
    @objc func Impuls(t: Timer)
    {
    	TaktImpulsAusführen()
    }
    
    /**
    Wird vom Timer im angegebenen Abstand aufgerufen.
    Wird für den benötigten Zweck überschrieben.
    */
    func TaktImpulsAusführen()
    {
        print("TaktImpulsAusführen")
    }

    /**
    Wird von der bei bei Touch / Mausklick aufgerufen.
    Wird für den benötigten Zweck überschrieben.
    - parameters:
    	- richtung: Die Himmelsrichtung, die dem Sektor des Klicks entspricht
    */
	func RichtungSetzen(richtung: Character)
    {
    	//print("RichtungSetzen: \(richtung)")
    }

    /**
    Wird von der bei bei Touch / Mausklick aufgerufen.
    Wird für den benötigten Zweck überschrieben.
    - parameters:
    	- rechts: Die Drehung, die der angeklicken Hälfte der View entspricht
    */
    func DrehungSetzen(rechts: Bool)
    {
    	//print("DrehungSetzen: \(rechts)")
    }
    
}

/**
Die Verwaltungsklasse der Schlange.
*/
class Schlange
{
    /** Darstellung des Kopfs. */
    private var kopf: KopfSymbol
    /** Darstellung des Rumpfs. */
    private var rumpf: [RumpfSymbol]
    
    /**
    Legt die Schlange an und stellt die Anfangsposition ein.
    - parameters:
    	- x: x-Position des Kopfs beim Start
    	- y: y-Position des Kopfs beim Start
    	- richtung: Bewegungsrichtung beim Start
    */
    init(x: Int, y: Int, richtung: Character)
    {
        kopf = KopfSymbol()
        rumpf = [RumpfSymbol]()
        for _ in 1...5
        {
            rumpf.append(RumpfSymbol())
        }
        kopf.AusrichtungSetzen(a: richtung)
        kopf.PositionSetzen(x: x, y: y)
        switch richtung
        {
            case "O":
                for i in 0..<rumpf.count
                {
                    rumpf[i].PositionSetzen(x: x - (rumpf.count - i), y: y)
                }
            case "N":
                for i in 0..<rumpf.count
                {
                    rumpf[i].PositionSetzen(x: x, y: y + (rumpf.count - i))
                }
            case "W":
                for i in 0..<rumpf.count
                {
                    rumpf[i].PositionSetzen(x: x + (rumpf.count - i), y: y)
                }
            case "S":
                for i in 0..<rumpf.count
                {
                    rumpf[i].PositionSetzen(x: x, y: y - (rumpf.count - i))
                }
            default:
            	fatalError("Falsche Ausrichtung im Konstruktor: \(richtung)")
		}
    }
    
    /**
    Bewegt die Schlange ein Feld in der eingestellten Richtung.
    - parameters:
    	- verlängern: wenn wahr, wird das letzte Rumpfelement nicht entfernt
    */
    func Bewegen(verlängern: Bool)
    {
        var x, y: Int
        var xAlt, yAlt: Int
        x = kopf.XPositionGeben()
        y = kopf.YPositionGeben()
        xAlt = x
        yAlt = y
        switch kopf.AusrichtungGeben()
        {
          case "O":
            x += 1
          case "N":
            y -= 1
          case "W":
            x -= 1
          case "S":
            y += 1
          default:
          	fatalError("Falsche Richtung: \(kopf.AusrichtungGeben()) in Schlange.Bewegen")
        }
        kopf.PositionSetzen(x: x, y: y)
        if !verlängern
        {
            rumpf[0].Entfernen()
            rumpf.remove(at: 0)
        }
        rumpf.append(RumpfSymbol())
        rumpf[rumpf.count - 1].PositionSetzen(x: xAlt, y: yAlt)
    }
    
    /**
    Setzt die Bewegungsrichtung der Schlange.
    - parameters:
    	- richtungNeu: Die (neue) Bewegungsrichtung
    */
    func RichtungSetzen(richtungNeu: Character)
    {
        kopf.AusrichtungSetzen(a: richtungNeu)
    }
    
    /**
    Gibt die x-Position des Schlangenkopfs zurück.
    - returns: x-Position des Schlangenkopfs
    */
    func XPositionGeben() -> Int
    {
        return kopf.XPositionGeben()
    }
    
    /**
    Gibt die y-Position des Schlangenkopfs zurück.
    - returns: y-Position des Schlangenkopfs
    */
    func YPositionGeben() -> Int
    {
        return kopf.YPositionGeben()
    }
        
    /**
    Prüft ob der Schlangenkopf auf der Position eines Rumpfsymbols ist, d.h. sich die Schlange selbst gebissen hat.
    - returns: wahr, wenn sich die Schlanger selbst gebissen hat, false sonst
    */
    func KopfInRumpf() -> Bool
    {
        return PunktInRumpf(x: kopf.XPositionGeben(), y: kopf.YPositionGeben())
    }

	/**
    Prüft, ob der angegebe Punkt auf der Position eines Rumpfsymbols ist.
    - parameters:
    	- x: die x-Koordinate des zu untersuchenden Punkts.
    	- y: die y-Koordinate des zu untersuchenden Punkts.
    - returns: wahr, wenn der Punkt zu einem Rumpfelement gehört.
    */
    func PunktInRumpf(x: Int, y: Int) -> Bool
    {
        for r in rumpf
        {
            if (x == r.XPositionGeben()) && (y == r.YPositionGeben())
            {
                return true
            }
        }
        return false
    }
}

/**
Rahmenklasse des Spiels.
*/
class Spiel: EreignisBehandlung
{
    /** Die Schlange. */
    private var schlange: Schlange
    /** Das Spielfeld. */
    private var spielfeldrand: RandSymbol
    /** Der Punktestand */
    private var punkte: Int
    /** Verlängerungsintervall */
    private let maxSchritte = 10
    /** Aktuelle Schrittzahl für Verlängerung */
    private var aktSchritte: Int
    /** Zufallsgeneror für das Spiel */
    private var zzGenerator: Zufall
    /** Restschritte für Sonderverlängerung der Schlange. */
    private var sonderZähler: Int
    /** Restschritte für Bremsen */
    private var bremsZähler: Int
    // Attribut zum Verwalten der Sonderfelder
    
    /**
    Baut die Basiselemente auf.
    */
    override init()
    {
        punkte = 0
        aktSchritte = maxSchritte
        sonderZähler = 0
        bremsZähler = 0
        spielfeldrand = RandSymbol()
        zzGenerator = Zufall()
        // Anlegen des Feldes der Sonderfelder
        let richtung: Character
        switch zzGenerator.ZufallGanz(top: 4)
        {
            case 0:
                richtung = "O"
            case 1:
                richtung = "N"
            case 2:
                richtung = "W"
            default:
                richtung = "S"
        }
        let xStart = -5 + zzGenerator.ZufallGanz(top: 10)
        let yStart = -4 + zzGenerator.ZufallGanz(top: 8)
        schlange = Schlange(x: xStart, y: yStart, richtung: richtung)
        super.init()
        StartTaktSetzen(starttakt: 500)
        // Anlegen der Sonderfelder
    }
    
    /**
    Erzeugt eine zufällige x-Koordinate auf den Spielfeld.
    - returns: zufällige x-Koordinate
    */
    func XKoordinateZufälligGeben() -> Int
    {
            return zzGenerator.ZufallGanz(top: spielfeldrand.XMaxGeben() - spielfeldrand.XMinGeben() + 1) + spielfeldrand.XMinGeben()
    }
    
    /**
    Erzeugt eine zufällige y-Koordinate auf den Spielfeld.
    - returns: zufällige y-Koordinate
     */
    func YKoordinateZufälligGeben() -> Int
    {
            return zzGenerator.ZufallGanz(top: spielfeldrand.YMaxGeben() - spielfeldrand.YMinGeben() + 1) + spielfeldrand.YMinGeben()
    }

    /**
    Testet, ob eine gegebene Position noch frei ist.
    - parameters:
    	- x: x-Koordinate der Position
    	- y: y-Koordinate der Position
    - returns: wahr, wenn die gegebene Position weder von der Schlange noch anderen Symbolen belegt ist.
    */
    func IstFrei(x: Int, y: Int) -> Bool
    {
    	//Berücksichtigung der Sonderfelder einbauen
        if (schlange.XPositionGeben() == x) && (schlange.YPositionGeben() == y)
        {
            return false
        }
        if schlange.PunktInRumpf(x: x, y: y)
        {
            return false
        }
        return true
    }
    
    /**
    Verändert den Punktestand um den angegebenen Wert.
	- parameters:
    	- delta: die Punktestandsänderung
    */
    func PunkteÄndern(delta: Int)
    {
            punkte += delta
    }
    
    /**
    Startet bzw. verlängert das Sonderwachstum der Schlange.
    */
    func Sonderwachstum()
    {
        sonderZähler += 10
    }
    
    /**
    Aktiviert die Verzögerung des Ablaufs.
    */
    func Bremsen()
    {
        bremsZähler = 100
        StartTaktSetzen(starttakt: 1000)
        spielfeldrand.SonderanzeigeSetzen(meldung: "+")
    }
    
    // Methode zum Entfernen eines Sonderfelds
    

    /**
    Bewegt die Schlange.
    */
    override func TaktImpulsAusführen()
    {
        if (schlange.XPositionGeben() <= spielfeldrand.XMaxGeben()) && (schlange.XPositionGeben() >= spielfeldrand.XMinGeben()) &&
           (schlange.YPositionGeben() <= spielfeldrand.YMaxGeben()) && (schlange.YPositionGeben() >= spielfeldrand.YMinGeben()) &&
           !schlange.KopfInRumpf()
        {
            aktSchritte -= 1
            if aktSchritte <= 0
            {
                schlange.Bewegen(verlängern: true)
                aktSchritte = maxSchritte
            }
            else
            {
                schlange.Bewegen(verlängern: false)
            }
            punkte += 1
            spielfeldrand.PunkteSetzen(punkte: punkte)
        }
        else
        {
            Anhalten()
            if schlange.KopfInRumpf()
            {
                spielfeldrand.EndemeldungSetzen(meldung: "Game Over - Schlange hat sich selbst gebissen")
            }
            else
            {
                spielfeldrand.EndemeldungSetzen(meldung: "Game Over - Schlange über Spielfeldrand")
            }
        }
    }
 
    /**
    Wertet die Mausklicks aus.
    
    Der erste Klick startet das Spiel
    - parameters:
    	- richtung: die Richtung, in die sich die Schlange bewegen soll
    */
    override func RichtungSetzen(richtung: Character)
    {
        schlange.RichtungSetzen(richtungNeu: richtung)
    }

    /**
    Wird von der Oberfläche bei bei Touch / Mausklick aufgerufen.

    Wird für den benötigten Zweck überschrieben.
    - parameters:
        - rechts: Die Drehung, die der angeklicken Hälfte der View entspricht
    */
    override func DrehungSetzen(rechts: Bool)
    {
    	return			//Leere Leistung, um die Ausgabe der Oberklassenfunktion zu verhindern.
    }
}




Spiel()
