/*:
## **Aufgabe Adventure (Teil 1)**
*/
import Foundation
#if os(macOS)
typealias View = NSView
typealias Button = NSButton
typealias TextView = NSTextView
typealias Label = NSTextField
typealias TableView = NSTableView
typealias TableViewDataSource = NSTableViewDataSource
typealias TableViewDelegate = NSTableViewDelegate
typealias Color = NSColor
import AppKit
#elseif os(iOS)
import UIKit
typealias View = UIView
typealias Button = UIButton
typealias TextView = UITextView
typealias Label = UILabel
typealias TableView = UITableView
typealias TableViewDataSource = UITableViewDataSource
typealias TableViewDelegate = UITableViewDelegate
typealias Color = UIColor
#endif
import PlaygroundSupport



/*
Konstanten für die Kartenanzeige und Gewölbegenerierung
*/
/** Anzahl der Räume */
let anzahlRäume = 40
/** Maximale Anzahl der Räume pro Zeile */
let breiteRaumRaster = 10
/** Maximale Anzahl der Räume pro Spalte */
let höheRaumRaster = 6
/** Raum als nicht besucht darstellen */
let nichtBesucht = 0
/** Raum als besucht darstellen */
let besucht = 1
/** Raum als bereits nach Ausgang durchsucht darstellen */
let ausgangGesucht = 2
/** Raum als markiert (Spieler anwesend) darstellen */
let markiert = 3



/**
Kommandos von Kontroller an Model

- author: Albert Wiedemann
- version: 1.0
*/
enum Kommandos
{
    case
    zimmerWechsel, schatzAktivieren, weiterKämpfen, kampfAbbrechen,
    schatzMitnehmen, schatzDalassen, handwaffeWeglegen,schutzwaffeWeglegen
}


/**
Der Kontroller verbindet die Programmlogik mit der Oberfläche.

- author: Albert Wiedemann
- version: 1.0
*/
class Kontroller
{
    /** Das Spiel, das kontrolliert wird */
    private var spiel: Spiel?

    /**
    Besetzt die Attribute
    */
    init()
    {
        spiel = nil
    }
    
    /**
	Trägt ein neues Kommado ein, falls nicht noch eines zur Verarbeitung ansteht.
	- parameters:
		- kdo: das einzutragende Kommando
		- param: der Kommandoparameter
     */
    func KommandoEintragen (kdo: Kommandos, param: Int)
    {
    	if spiel != nil
    	{
    		if spiel!.spieler.KommandosAusführen(kdo: kdo, param: param)
        	{
                spiel!.anzeige.EndemeldungAnzeigen()
            }
        }
    }
    
    /** Setzt das kontrollierte Spiel
    - parameters:
    	- s: das Spiel
    */
    func SpielSetzen (s: Spiel)
    {
    	spiel = s
    }
}

/**
Raumdaten für die Landkarte.

- author: Albert Wiedemann
- version: 1.0
*/
class LandkarteneintragRaum
{
    /** Raumnummer */
    let nummer: Int
    /** x-Koordinate */
    let x: Int
    /** y-Koordinate */
    let y: Int
    /** Darstellungsmodus */
    let mode: Int

    /**
	Besetzt die Raumdaten
	- parameters:
		- nummer: Raumnummer
		- x: x-Koordinate
		- y: y-Koordinate
		- mode: Darstellungsmodus
     */
    init(nummer: Int, x: Int, y: Int, mode: Int)
    {
        self.nummer = nummer
        self.x = x
        self.y = y
        self.mode = mode
    }
}

/**
Die Daten eines Wegs für die Landkarte.

- author: Albert Wiedemann
- version: 1.0
*/
class LandkarteneintragWeg
{
    /** x-Koordinate des Startpunkts */
    var x1: Int
    /** y-Koordinate des Startpunkts */
    var y1: Int
    /** x-Koordinate des Zielpunkts */
    var x2: Int
    /** y-Koordinate des Zielpunkts */
    var y2: Int

    /**
	Besetzt die Wegdaten
	- parameters:
		- x1: x-Koordinate des Startpunkts
		- y1: y-Koordinate des Startpunkts
		- x2: x-Koordinate des Zielpunkts
		- y2: y-Koordinate des Zielpunkts
     */
    init(x1: Int, y1: Int, x2: Int, y2: Int)
    {
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
    }
}

/**
Grafische Darstellung der Landkarte

- author: Albert Wiedemann
- version: 1.0
*/
class Landkarte: View
{
    /** Höhe für die Darstellung eines Raums. */
    let raumHöhe = 20
    /** Breite für die Darstellung eines Raums. */
    let raumBreite = 20
    /** Abstand zweier Räume in x-Richtung */
    let abstandX = 40
    /** Abstand zweier Räume in y-Richtung */
    let abstandY = 40
    /** Oberer Rand der Zeichnung */
    let karteOben = 5
    /** Unterer Rand der Zeichnung */
    let karteUnten: Int
    /** Linker Rand der Zeichnung */
    let karteLinks = 5
    /** Rechter Rand der Zeichnung */
    let karteRechts: Int
    /** Daten der Räume */
    var räume: [LandkarteneintragRaum]
    /** Daten der Wege */
   	var wege: [LandkarteneintragWeg]
    /** Raum, in dem die Maus gedrückt wurde oder -1 */
    var raumGeklickt: Int
    /** Der Kontroller für die Verbindung zur Programmlogik */
    let kontroller: Kontroller

    /**
    Legt die Datenstrukturen zum Speichern der Räume und Wege an
	- parameters:
		- k: der Kontroller für das Eintragen der Kommandos
    */
    init(k: Kontroller)
    {
        karteUnten = karteOben + (raumHöhe *  höheRaumRaster + abstandY * (höheRaumRaster - 1))
        karteRechts = karteLinks + raumBreite *  breiteRaumRaster + abstandX * (breiteRaumRaster - 1)
        räume = [LandkarteneintragRaum]()
        wege =  [LandkarteneintragWeg]()
        raumGeklickt = -1
        kontroller = k
        super.init(frame: CGRect(x: 0, y: 0, width: karteLinks + karteRechts, height: karteOben + karteUnten))
    }
    
    /**
	Dummy
    */
    required init?(coder: NSCoder)
    {
    	return nil
    }
    
    #if os(macOS)
    /**
    Aufgerufen, wenn die Maus gedrückt wird.
    - parameters:
        - event: das Ereignis, zu dem der Klick gehört
    */
    override func mouseDown(with event: NSEvent)
    {
        var p = convert(event.locationInWindow, from: nil)
        p.y = bounds.height - p.y
        MausGedrückt(p: p)
    }

    /**
    Aufgerufen, wenn die Maus losgelassen wird.
    - parameters:
        - event: das Ereignis, zu dem der Klick gehört
    */
    override func mouseUp(with event: NSEvent)
    {
        var p = convert(event.locationInWindow, from: nil)
        p.y = bounds.height - p.y
        MausLosgelassen(p: p)
    }
    
    #elseif os(iOS)
    /**
    Aufgerufen, wenn ein oder mehr Finger die Tastatur berühren.
    - parameters:
        - touches: Menge an Berüchrungsbeschreibungen für den Beginn des Ereignisses
        - event: das Gesamtereignis, zu dem die Berührungen gehören
    */
    override func touchesBegan(_ touches: Set<UITouch>, with event: UIEvent?) {
        for t in touches as Set<UITouch>
        {
            MausGedrückt (p: t.location(in: self))
        }
        super.touchesBegan(touches, with: event)
    }

    /**
    Aufgerufen, wenn ein oder mehr Finger die Tastatur verlassen.
    - parameters:
        - touches: Menge an Berührungsbeschreibungen für den Beginn des Ereignisses
        - event: das Gesamtereignis, zu dem die Berührungen gehören
    */
    override func touchesEnded(_ touches: Set<UITouch>, with event: UIEvent?)
    {
        for t in touches as Set<UITouch>
        {
            MausLosgelassen(p: t.location(in: self))
        }
        super.touchesEnded(touches, with: event)
    }

    override func tintColorDidChange()
    {
        setNeedsDisplay()
    }
    #endif

    /**
	Sucht, ob es zum angegeben Punkt einen Raum gibt
	- parameters:
		- p: der Punkt, an dem die Maus gedrückt wurde
	- returns: Raumnummer oder -1 für "kein Raum gefunden"
    */
	private func RaumSuchen(p: CGPoint) -> Int
    {
        for raum in räume
        {
            let r = CGRect(x: karteLinks + (raumBreite + abstandX) *  raum.x, y: karteOben + (raumHöhe + abstandY) *  (höheRaumRaster - 1 - raum.y), width: raumBreite, height: raumHöhe)
            if r.contains(p)
            {
                return raum.nummer
            }
        }
        return -1
    }
    
    /**
	Wird beim Drücken der Maus aufgerufen und bestimmt den (gegebenenfalls) gedrückten Raum
	- parameters:
		- p: der Klickpunkt
    */
    func MausGedrückt (p: CGPoint)
    {
        raumGeklickt = RaumSuchen(p: p)
    }
    
    /**
    Wird beim Loslassen der Maus aufgerufen und bestimmt den (gegebenenfalls) gedrückten Raum
	- parameters:
		- p: der Klickpunkt
    */
    func MausLosgelassen (p: CGPoint)
    {
        if (raumGeklickt == RaumSuchen(p: p)) && (raumGeklickt != -1)
        {
            kontroller.KommandoEintragen(kdo: Kommandos.zimmerWechsel, param: raumGeklickt);
        }
        raumGeklickt = -1
    }

    /**
	Zeichnet einen Raum der Landkarte
	- parameters:
		- raum: die Raumdaten
		- g: das Grafikobjekt
    */
    func RaumZeichnen(raum: LandkarteneintragRaum, g: CGContext)
    {
        g.saveGState()
        let oben = karteOben + (raumHöhe + abstandY) *  (höheRaumRaster - 1 - raum.y);
        let links = karteLinks + (raumBreite + abstandX) *  raum.x;
        switch raum.mode
        {
            case nichtBesucht:
                g.setFillColor((Color.lightGray.cgColor))
            case besucht:
                g.setFillColor((Color.yellow.cgColor))
            case ausgangGesucht:
                g.setFillColor((Color.green.cgColor))
            case markiert:
                g.setFillColor((Color.red.cgColor))
            default:
                g.setFillColor((Color.white.cgColor))
        }
        g.fill(CGRect(x: links, y: oben, width: raumBreite, height: raumHöhe))
        g.restoreGState()
        g.stroke(CGRect(x: links, y: oben, width: raumBreite, height: raumHöhe))
        let font = CTFontCreateWithName (("Helvetica" as CFString?)!, 10, nil)
        let attrString = NSAttributedString(string: "\(raum.nummer)", attributes: [(NSAttributedString.Key(rawValue: kCTForegroundColorAttributeName as String as String)):Color.black.cgColor, (NSAttributedString.Key(rawValue: kCTFontAttributeName as String as String)):font])
        let line = CTLineCreateWithAttributedString(attrString)
        g.textPosition = CGPoint(x: links + 2, y: oben + 13)
        CTLineDraw(line, g)
    }
    
    /**
	Zeichnet einen Weg der Landkarte
	- parameters:
		- x1: x-Koordiante des Weganfangs
		- y1: y-Koordiante des Weganfangs
		- x2: x-Koordiante des Wegendes
		- y2: y-Koordiante des Wegendes
		- g: das Grafikobjekt
     */
    func WegZeichnen (x1: Int, y1: Int, x2: Int, y2: Int, g: CGContext)
    {
        g.saveGState()
        g.setStrokeColor(Color.yellow.cgColor)
        g.setLineWidth(3)
        g.beginPath()
        if x1 == x2
        {
            let x = karteLinks + (raumBreite + abstandX) *  x1 - 1 + raumBreite / 2
            let yu = karteOben + (raumHöhe + abstandY) *  (höheRaumRaster - 1 - y1)
            let yo = karteOben + (raumHöhe + abstandY) *  (höheRaumRaster - 1 - y2) + raumHöhe
            g.move(to: CGPoint(x: x, y: yo))
            g.addLine(to: CGPoint(x: x, y: yu))
            g.strokePath()
            g.setStrokeColor(Color.black.cgColor)
            g.setLineWidth(1)
            g.move(to: CGPoint(x: x-2, y: yo))
            g.addLine(to: CGPoint(x: x-2, y: yu))
            g.move(to: CGPoint(x: x+2, y: yo))
            g.addLine(to: CGPoint(x: x+2, y: yu))
        }
        else if (y1 == y2)
        {
            let xl = karteLinks + (raumBreite + abstandX) *  x1 + raumBreite
            let xr = karteLinks + (raumBreite + abstandX) *  x2
            let y = karteOben + (raumHöhe + abstandY) * (höheRaumRaster - 1 - y1) + raumHöhe / 2
            g.move(to: CGPoint(x: xl, y: y))
            g.addLine(to: CGPoint(x: xr, y: y))
            g.strokePath()
            g.setStrokeColor(Color.black.cgColor)
            g.setLineWidth(1)
            g.move(to: CGPoint(x: xl, y: y-2))
            g.addLine(to: CGPoint(x: xr, y: y-2))
            g.move(to: CGPoint(x: xl, y: y+2))
            g.addLine(to: CGPoint(x: xr, y: y+2))
        }
        else if (y1 < y2)
        {
            let xl = karteLinks + (raumBreite + abstandX) *  x1 + raumBreite
            let xr = karteLinks + (raumBreite + abstandX) *  x2
            let yu = karteOben + (raumHöhe + abstandY) *  (höheRaumRaster - 1 - y1)
            let yo = karteOben + (raumHöhe + abstandY) *  (höheRaumRaster - 1 - y2) + raumHöhe
            g.move(to: CGPoint(x: xl, y: yu))
            g.addLine(to: CGPoint(x: xr, y: yo))
            g.strokePath()
            g.setStrokeColor(Color.black.cgColor)
            g.setLineWidth(1)
            g.move(to: CGPoint(x: xl, y: yu+2))
            g.addLine(to: CGPoint(x: xr+2, y: yo))
            g.move(to: CGPoint(x: xl-2, y: yu))
            g.addLine(to: CGPoint(x: xr, y: yo-2))
        }
        else
        {
            let xl = karteLinks + (raumBreite + abstandX) *  x1 + raumBreite
            let xr = karteLinks + (raumBreite + abstandX) *  x2
            let yu = karteOben + (raumHöhe + abstandY) *  (höheRaumRaster - 1 - y2)
            let yo = karteOben + (raumHöhe + abstandY) *  (höheRaumRaster - 1 - y1) + raumHöhe
            g.move(to: CGPoint(x: xl, y: yo))
            g.addLine(to: CGPoint(x: xr, y: yu))
            g.strokePath()
            g.setStrokeColor(Color.black.cgColor)
            g.setLineWidth(1)
            g.move(to: CGPoint(x: xl, y: yo-2))
            g.addLine(to: CGPoint(x: xr+2, y: yu))
            g.move(to: CGPoint(x: xl-2, y: yo))
            g.addLine(to: CGPoint(x: xr, y: yu+2))
        }
        g.strokePath()
        g.restoreGState()
    }
    
    /**
    Zeichnet die Landkarte
    - parameters:
        - rect: das (mindestens) neu zu zeichnende Rechteck
    */
    override func draw(_ rect: CGRect)
    {
        #if os(macOS)
        if let g = NSGraphicsContext.current?.cgContext
        {
            g.scaleBy(x:1, y:-1)
            g.translateBy(x:0, y:-self.bounds.height)
            g.textMatrix = CGAffineTransform(a: 1, b: 0.0, c: 0.0, d: -1, tx: 0.0, ty: 0.0)
            g.setFillColor(Color.white.cgColor)
            g.fill(CGRect(x: karteLinks - 2, y: karteOben - 2, width: karteRechts - karteLinks + 3, height: karteUnten - karteOben + 3))
            g.setStrokeColor(Color.black.cgColor)
            g.stroke(CGRect(x: karteLinks - 2, y: karteOben - 2, width: karteRechts - karteLinks + 3, height: karteUnten - karteOben + 3))
            for raum in räume
            {
                RaumZeichnen(raum: raum, g: g)
            }
            for weg in wege
            {
                WegZeichnen(x1: weg.x1, y1: weg.y1, x2: weg.x2, y2: weg.y2, g: g)
            }
        }
        #elseif os(iOS)
        if let g = UIGraphicsGetCurrentContext()
        {
            g.textMatrix  = CGAffineTransform(a: 1, b: 0.0, c: 0.0, d: -1, tx: 0.0, ty: 0.0)
            g.setFillColor(Color.white.cgColor)
            g.fill(CGRect(x: karteLinks - 2, y: karteOben - 2, width: karteRechts - karteLinks + 3, height: karteUnten - karteOben + 3))
            g.setStrokeColor(Color.black.cgColor)
            g.stroke(CGRect(x: karteLinks - 2, y: karteOben - 2, width: karteRechts - karteLinks + 3, height: karteUnten - karteOben + 3))
            for raum in räume
            {
                RaumZeichnen(raum: raum, g: g)
            }
            for weg in wege
            {
                WegZeichnen(x1: weg.x1, y1: weg.y1, x2: weg.x2, y2: weg.y2, g: g)
            }
        }
        #endif
    }
    
    /**
    Löscht die gespeicherten Landkartendaten
    */
    func LandkartendatenLöschen()
    {
        räume.removeAll()
        wege.removeAll()
    }
    
    /**
	Trägt die Daten für einen weiteren Raum ein
	- parameters:
		- r: Raumdaten
     */
    func RaumEintragen(r: LandkarteneintragRaum)
    {
        for raum in räume
        {
            if r.nummer == raum.nummer
            {
                return
            }
        }
        räume.append(r)
    }
    
    /**
	Trägt die Daten für einen weiteren Weg ein
	- parameters:
		- w: Wegdaten
     */
    func WegEintragen(w: LandkarteneintragWeg)
    {
        var hilf: Int
        if w.x1 > w.x2
        {
            hilf = w.x2
            w.x2 = w.x1
            w.x1 = hilf
            hilf = w.y2
            w.y2 = w.y1
            w.y1 = hilf
        }
        else if (w.x1 == w.x2) && (w.y1 > w.y2)
        {
            hilf = w.y2
            w.y2 = w.y1
            w.y1 = hilf
        }
        for weg in wege
        {
            if (w.x1 == weg.x1) && (w.x2 == weg.x2) && (w.y1 == weg.y1) && (w.y2 == weg.y2)
            {
                return
            }
        }
        wege.append(w)
    }
}

/**
Zeigt die Spielregel an

- author: Albert Wiedemann
- version: 1.0
*/
class SpielregelAnzeige: TextView
{
    private let t =
            "Du bist ein Zauberer und in einem unterirdischen Gewölbe aus Räumen und Verbindungsgängen gefangen.\n" +
            "Um das Gewölbe verlassen zu können, musst du drei magische Schlüssel finden;\n" +
            "dann kannst du den Ausgang finden und ins Freie gelangen.\n" +
            "Damit das Spiel nicht zu leicht wird, hausen in den meisten Räumen Monster, die es zu besiegen gilt.\n" +
            "Dazu findest du in den Räumen Waffen und magische Artefakte, die Zaubersprüche.\n" +
            "Die Waffen unterteilen sich in Angriffs- und Schutzwaffen. Du kannst eine deiner Angriffswaffen\n" +
            "in die Hand nehmen, in dem du sie in deiner Schatzübersicht auswählst und dann den Knopf 'Aktivieren' drückst.\n" +
            "Genau so kannst du eine deiner Schutzwaffen anziehen.\n" +
            "Einen Zauberspruch kannst du anwenden, in dem du ihn in der Übersicht deiner Schätze auswählst\n" +
            "und dann ebenfalls den Knopf 'Aktivieren' drückst." +
            "\n" +
            "Wenn du einen neuen Raum betrittst (Klick in der Landkarte) muss du zuerst mit dem dort\n" +
            "befindlichen Monster kämpfen. Dafür zählen vor allem deine Handwaffe und die Schutzwaffe,\n" +
            "die du angezogen hast. Wenn du gewinnst, ist alles gut. Wenn du verlierst, kannst du den\n" +
            "Kampf abbrechen (dann hast du verloren) oder Fortsetzen. Vor den Fortsetzen kannst du eine\n" +
            "neue Handwaffe und eine neue Schutzwaffe wählen. Du kannst auch einen Zauberspruch ausführen,\n" +
            "von dem du vermutest, dass er beim Kampf hilft." +
            "\n" +
            "Wenn du den Kampf gewonnen hast, kannst du dich an den Schätzen im Raum bedienen. Du wählst ihn\n" +
            "dazu in der Schatzübersicht des Raums aus und drückst auf 'Mitnehmen'. Einzige Einschränkung\n" +
            "ist, dass jeder Schatz ein Gewicht hat und du nur ein bestimmtes Maximalgewicht tragen kannst.\n" +
            "Auch belasten die Zaubersprüche dein Gehirn - sie haben ein Mentalgewicht. Auch die Kapazität\n" +
            "deines Gehirns ist begrenzt und lässt nicht zu viele magische Schätze zu. Damit dich früher\n" +
            "gefundene, schwächere Schätze nicht daran hindern, neue, bessere Schätze mitzunehmen, kannst\n" +
            "du auch Schätze in einem Raum ablegen - du wählst einen deiner Schätze\n" +
            "in der Übersicht deiner Schätze aus und drückst 'Dalassen'.\n" +
            "Jetzt weißt du alles, was du Wissen musst, den Rest musst du selbst heraus finden." +
            "\n" +
            "\n" +
            "Viel Spass :-)"

    /**
    Legt die Komponeten an
    */
    func SpielregelAnzeigen()
    {
		#if os(macOS)
    	string = t
    	#elseif os(iOS)
    	text = t
    	#endif
    }
}

/**
Beschreibt die in der Anzeige vorhandenen Setze-Methoden.

- author: Albert Wiedemann
- version: 1.0
*/
protocol AnzeigeInterface
{
    /**
	Setzt den Namen der Handwaffe
	- parameters:
		- name: Name der Handwaffe
     */
    func HandwaffenNameSetzen(name: String)

    /**
	Setzt den Namen der Schutzwaffe
	- parameters:
		- name Name der Schutzwaffe
     */
    func SchutzwaffenNameSetzen(name: String)

    /**
    Setzt die Stärkenanzeige
	- parameters:
		- st: Wert der Stärke
    */
    func StärkeSetzen(st: Int)

    /**
    Setzt die Anzahl der Schlüssel
	- parameters:
		- anz: Anzahl
    */
    func SchlüsselAnzahlSetzen(anz: Int)
    
    /**
    Setzt das Gesamtgewicht der Schätze
	- parameters:
		- gewicht: Gesamtgewicht
    */
    func GesamtgewichtSetzen(gewicht: Double)

    /**
    Setzt die Liste der Spielerschätze
	- parameters:
		- texte: Namen der Schätze
    */
    func SpielerSchätzeSetzen(texte: [String])
    
    /**
    Setzt die Liste der Raumschätze
	- parameters:
		- texte: Namen der Schätze
    */
    func RaumSchätzeSetzen(texte: [String])

    /**
    Trägt den Monsternamen ein und blendet die nötigen Anzeigen für den Monsterkampf ein / aus.
	- parameters:
		- text der Monstername
    */
    func RundeVerlorenSetzen(text: String)

    /**
    Beendet den Monsterkampf und blendet die nötigen Anzeigen ein / aus.
    */
    func KampfZuEndeSetzen()
    
    /**
    Fügt eine neue Botschaft an die Anzeige an.
	- parameters:
		- text Botschaft
    */
    func NeueBotschaftEintragen(text: String)

    /**
    Löscht die Daten der Landkartenanzeige
    */
    func LandkartendatenLöschen()
    
    /**
    Trägt neue Rauminformation in die Landkartenanzeige ein.
	- parameters:
		- r: der einzutragende Raum
    */
    func RaumEintragen(r: LandkarteneintragRaum)
    
    /**
    Trägt neue Weginformation in die Landkartenanzeige ein.
	- parameters:
		- w: der einzutragende Weg
    */
    func WegEintragen(w: LandkarteneintragWeg)

    /**
    Veranlasst das Neuzeichnen der Landkartenanzeige
    */
    func LandkarteNeuZeichnen()
    
    /**
    Zeigt die Meldung über das erfolgreiche Spielende an.
    */
    func EndemeldungAnzeigen()
}

/**
Beschreibt die vollständige Anzeige des Spiels

- author: Albert Wiedemann
- version: 1.0
*/
class Anzeige: View, TableViewDataSource, AnzeigeInterface, TableViewDelegate
{
    /** Fenster und Oberflächenelemente */
    private var hauptPanel: View
    private var mapping: Landkarte
    private var stärkeText: Label
    private var handwaffenText: Label
    private var kampfMonster: Label
    private var infoText: TextView
    private var schlüsselAnzahl: Label
    private var schutzwaffenText: Label
    private var raumListe: TableView
    private var schatzListe: TableView
    private var schatzText: Label
    private var raumText: Label
    private var gesamtGewicht: Label
    private var kampfItems: View
    private var raumItems: View
    private var endeMeldung: View
    private var japadapadu: Label
    private var gefunden: Label
    private var handwaffeWeglegen: Button
    private var schutzwaffeWeglegen: Button
    private var spielregelAnzeigen: Button
    private var mitnehmen: Button
    private var dalassen: Button
    private var aktivieren: Button
    private var abbrechen: Button
    private var weiterkämpfen: Button
    private var spielregel: View
    private var spielregelAnzeige: SpielregelAnzeige
    private var schließenKnopf: Button
    private var raumlistenDaten: [String]
    private var schatzlistenDaten: [String]
    #if os(macOS)
    private var schatzlistenScroll: NSScrollView
    private var raumlistenScroll: NSScrollView
    private var infoTextScroll: NSScrollView
    #endif
    
    /** der Kontroller für die Anzeige */
    private var kontroller: Kontroller

    /**
	Speicher den Kontroller und legt die Elemente an.
	- parameters:
		- k: der kontroller für die Verbindung zur Programmlogik
    */
    init(k: Kontroller)
    {
        kontroller = k;
        #if os(macOS)
        hauptPanel = View(frame: CGRect(x: 0, y: 600 - 30 - 570, width: 800, height: 570))
        mapping = Landkarte(k: kontroller)
        mapping.frame.origin = CGPoint(x: 10, y: 570 - 10 - mapping.bounds.height)
        schatzListe = TableView(frame: CGRect(x: 0, y: 0, width: 180, height: 140))
        schatzlistenDaten = [String]()
        schatzlistenScroll = NSScrollView(frame: CGRect(x: 610, y: 570 - 200 - 140, width: 180, height: 140))
        schatzlistenScroll.documentView = schatzListe
        stärkeText = Label(frame: CGRect(x: 420, y: 570 - 350 - 20, width: 100, height: 20))
        handwaffenText = Label(frame: CGRect(x: 420, y: 570 - 380 - 20, width: 160, height: 20))
        kampfMonster = Label(frame: CGRect(x: 20, y: 330 - 30 - 20, width: 250, height: 20))
        infoText = TextView(frame: CGRect(x: 0, y: 0, width: 380, height: 140))
        infoTextScroll = NSScrollView(frame: CGRect(x: 20, y: 570 - 350 - 140, width: 380, height: 140))
        infoTextScroll.documentView = infoText
        schlüsselAnzahl = Label(frame: CGRect(x: 420, y: 570 - 440 - 20, width: 100, height: 20))
        schutzwaffenText = Label(frame: CGRect(x: 420, y: 570 - 410 - 20, width: 160, height: 20))
        raumListe = TableView(frame: CGRect(x: 0, y: 0, width: 180, height: 110))
        raumlistenDaten = [String]()
        raumlistenScroll = NSScrollView(frame: CGRect(x: 10, y: 180 - 35 - 110, width: 180, height: 110))
        raumlistenScroll.documentView = raumListe
        schatzText = Label(frame: CGRect(x: 610, y: 570 - 180 - 20, width: 180, height: 20))
        raumText = Label(frame: CGRect(x: 10, y: 180 - 15 - 20, width: 180, height: 20))
        gesamtGewicht = Label(frame: CGRect(x: 420, y: 570 - 470 - 20, width: 170, height: 20))
        kampfItems = View(frame: CGRect(x: 10, y: 570 - 10 - 330, width: 570, height: 330))
        raumItems = View(frame: CGRect(x: 600, y: 570 - 0 - 180, width: 200, height: 180))
        endeMeldung = View(frame: CGRect(x: 0, y: 570 - 0 - 500, width: 800, height: 500))
        japadapadu = Label(frame: CGRect(x: 100, y: 570 - 100 - 20, width: 200, height: 20))
        gefunden = Label(frame: CGRect(x: 100, y: 570 - 130 - 20, width: 200, height: 20))
        spielregel = View(frame: CGRect(x: 0, y: 570 - 0 - 500, width: 800, height: 500))
        spielregelAnzeige = SpielregelAnzeige(frame: CGRect(x: 0, y: 570 - 0 - 450, width: 800, height: 450))
        handwaffeWeglegen = Button(title: "Handwaffe weglegen", target: nil, action: #selector(Anzeige.KnopfGedrückt))
        schutzwaffeWeglegen = Button(title: "Schutzwaffe weglegen", target: nil, action: #selector(Anzeige.KnopfGedrückt))
        spielregelAnzeigen = Button(title: "Spielregel", target: nil, action: #selector(Anzeige.KnopfGedrückt))
        mitnehmen = Button(title: "Mitnehmen", target: nil, action: #selector(Anzeige.KnopfGedrückt))
        dalassen = Button(title: "Dalassen", target: nil, action: #selector(Anzeige.KnopfGedrückt))
        aktivieren = Button(title: "Aktivieren", target: nil, action: #selector(Anzeige.KnopfGedrückt))
        abbrechen = Button(title: "Kampf abbrechen", target: nil, action: #selector(Anzeige.KnopfGedrückt))
        weiterkämpfen = Button(title: "Weiter kämpfen", target: nil, action: #selector(Anzeige.KnopfGedrückt))
        schließenKnopf = Button(title: "Schließen", target: nil, action: #selector(Anzeige.KnopfGedrückt))
        aktivieren.frame = CGRect(x: 605, y: 570 - 340 - 30, width: 90, height: 30)
        abbrechen.frame = CGRect(x: 200, y: 570 - 60 - 30, width: 150, height: 30)
        weiterkämpfen.frame = CGRect(x: 20, y: 570 - 60 - 30, width: 150, height: 30)
        schließenKnopf.frame = CGRect(x: 300, y: 570 - 460 - 30, width: 200, height: 30)
        handwaffeWeglegen.frame =  CGRect(x: 600, y: 570 - 370 - 30, width: 190, height: 30)
        schutzwaffeWeglegen.frame = CGRect(x: 600, y: 570 - 400 - 30, width: 190, height: 30)
        spielregelAnzeigen.frame = CGRect(x: 600, y: 570 - 460 - 30, width: 190, height: 30)
        mitnehmen.frame = CGRect(x: 5, y: 180 - 145 - 30, width: 190, height: 30)
        dalassen.frame = CGRect(x: 705, y: 570 - 340 - 30, width: 90, height: 30)
        aktivieren.frame = CGRect(x: 605, y: 570 - 340 - 30, width: 90, height: 30)
        abbrechen.frame = CGRect(x: 200, y: 330 - 60 - 30, width: 150, height: 30)
        weiterkämpfen.frame = CGRect(x: 20, y: 330 - 60 - 30, width: 150, height: 30)
        #elseif os(iOS)
        hauptPanel = View(frame: CGRect(x: 0, y: 0, width: 800, height: 570))
        mapping = Landkarte(k: kontroller)
        mapping.frame.origin = CGPoint(x: 10, y: 10)
        schatzListe = TableView(frame: CGRect(x: 610, y: 200, width: 180, height: 140))
        schatzlistenDaten = [String]()
        stärkeText = Label(frame: CGRect(x: 420, y: 350, width: 100, height: 20))
        handwaffenText = Label(frame: CGRect(x: 420, y: 380, width: 160, height: 20))
        kampfMonster = Label(frame: CGRect(x: 20, y: 30, width: 250, height: 20))
        infoText = TextView(frame: CGRect(x: 20, y: 350, width: 380, height: 140))
        schlüsselAnzahl = Label(frame: CGRect(x: 420, y: 440, width: 100, height: 20))
        schutzwaffenText = Label(frame: CGRect(x: 420, y: 410, width: 160, height: 20))
        raumListe = TableView(frame: CGRect(x: 10, y: 35, width: 180, height: 110))
        raumlistenDaten = [String]()
        schatzText = Label(frame: CGRect(x: 610, y: 180, width: 180, height: 20))
        raumText = Label(frame: CGRect(x: 10, y: 15, width: 180, height: 20))
        gesamtGewicht = Label(frame: CGRect(x: 420, y: 470, width: 170, height: 20))
        kampfItems = View(frame: CGRect(x: 10, y: 10, width: 570, height: 330))
        raumItems = View(frame: CGRect(x: 600, y: 0, width: 200, height: 180))
        endeMeldung = View(frame: CGRect(x: 0, y: 0, width: 800, height: 500))
        japadapadu = Label(frame: CGRect(x: 100, y: 100, width: 200, height: 20))
        gefunden = Label(frame: CGRect(x: 100, y: 130, width: 200, height: 20))
        spielregel = View(frame: CGRect(x: 0, y: 0, width: 800, height: 500))
        spielregelAnzeige = SpielregelAnzeige(frame: CGRect(x: 0, y: 0, width: 800, height: 450))
        handwaffeWeglegen = Button(type: .roundedRect)
        schutzwaffeWeglegen = Button(type: .roundedRect)
        spielregelAnzeigen = Button(type: .roundedRect)
        mitnehmen = Button(type: .roundedRect)
        dalassen = Button(type: .roundedRect)
        aktivieren = Button(type: .roundedRect)
        abbrechen = Button(type: .roundedRect)
        weiterkämpfen = Button(type: .roundedRect)
        schließenKnopf = Button(type: .roundedRect)
        aktivieren.frame = CGRect(x: 605, y: 340, width: 90, height: 30)
        abbrechen.frame = CGRect(x: 200, y: 60, width: 150, height: 30)
        weiterkämpfen.frame = CGRect(x: 20, y: 60, width: 150, height: 30)
        schließenKnopf.frame = CGRect(x: 300, y: 460, width: 200, height: 30)
        handwaffeWeglegen.frame =  CGRect(x: 600, y: 370, width: 190, height: 30)
        schutzwaffeWeglegen.frame = CGRect(x: 600, y: 400, width: 190, height: 30)
        spielregelAnzeigen.frame = CGRect(x: 600, y: 460, width: 190, height: 30)
        mitnehmen.frame = CGRect(x: 5, y: 145, width: 190, height: 30)
        dalassen.frame = CGRect(x: 705, y: 340, width: 90, height: 30)
        aktivieren.frame = CGRect(x: 605, y: 340, width: 90, height: 30)
        abbrechen.frame = CGRect(x: 200, y: 60, width: 150, height: 30)
        weiterkämpfen.frame = CGRect(x: 20, y: 60, width: 150, height: 30)
        #endif
        super.init(frame: CGRect(x: 0, y: 0, width: 800, height: 600))
        KomponentenAnlegen()
    }
    
    /**
	Dummy
    */
    required init?(coder: NSCoder)
    {
    	return nil
    }
    
    
    #if os(macOS)
    /**
	Meldet die Zeilenanzahl der Tabelle
	- parameters:
		- tableView: die View, die anfragt
	- returns: Zeilenanzahl
    */
    func numberOfRows(in tableView: NSTableView) -> Int
    {
		if tableView === schatzListe
    	{
    		return schatzlistenDaten.count
        }
		else if tableView === raumListe
    	{
    		return raumlistenDaten.count
        }
        else
        {
        	return 0
        }
    }

    /**
	Erzeugt den Inhalt für die gewünschte Zeile
	- parameters:
		- tableView: die View, die anfragt
		- tableColumn: die gewünschte Spalte
		- row: die gewünschte Zeile
	- returns: Zelle
    */
	func tableView(_ tableView: NSTableView, viewFor tableColumn: NSTableColumn?, row: Int) -> NSView?
	{
        var t: String
        if tableView === schatzListe
    	{
    		t = schatzlistenDaten[row]
        }
		else if tableView === raumListe
    	{
    		t = raumlistenDaten[row]
        }
        else
        {
        	t = "---"
        }
        
        let cellView = NSTextField(string: t)
        cellView.identifier = NSUserInterfaceItemIdentifier(rawValue: "Zeile")
        cellView.isEditable = false
        cellView.isBordered = false
        cellView.isEnabled = true
        cellView.isSelectable = false
        return cellView
    }
    
    func tableView(_ tableView: NSTableView, heightOfRow row: Int) -> CGFloat
    {
        return 20
    }
    
    func tableView(_ tableView: NSTableView, shouldSelectRow row: Int) -> Bool
    {
        return true
    }
    
    #elseif os(iOS)
    /**
	Meldet die Zeilenanzahl der Tabelle
	- parameters:
		- tableView: die View, die anfragt
		- section: Anzshl der Sectionen
	- returns: Zeilenanzahl
    */
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int
    {
		if tableView === schatzListe
    	{
    		return schatzlistenDaten.count
        }
		else if tableView === raumListe
    	{
    		return raumlistenDaten.count
        }
        else
        {
        	return 0
        }
    }

    /**
	Erzeugt die Zelle für die gewünschte Zeile
	- parameters:
		- tableView: die View, die anfragt
		- indexPath: Angezeigt anzahl
	- returns: Zelle
    */
    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
       // Fetch a cell of the appropriate type.
       let cell = tableView.dequeueReusableCell(withIdentifier: "Zeile", for: indexPath)
       
       // Configure the cell’s contents.
		if tableView === schatzListe
    	{
    		cell.textLabel!.text = schatzlistenDaten[indexPath.row]
        }
		else if tableView === raumListe
    	{
    		cell.textLabel!.text = raumlistenDaten[indexPath.row]
        }
        else
        {
        	cell.textLabel!.text = "---"
        }
       return cell
    }
    #endif

    /**
    Legt die Komponenten der Oberfläche an
    */
    func KomponentenAnlegen()
    {
		schatzListe.dataSource = self
    	schatzListe.rowHeight = 20
    	schatzListe.allowsMultipleSelection = false
    	#if os(macOS)
    	schatzListe.headerView = nil
    	schatzListe.delegate = self
    	schatzListe.usesAutomaticRowHeights
        schatzListe.isEnabled = true
    	let col = NSTableColumn(identifier: NSUserInterfaceItemIdentifier(rawValue: "Zeile") )
		col.width = 180
		schatzListe.addTableColumn(col)
    	#elseif os(iOS)
    	schatzListe.allowsSelection = true
        schatzListe.register(UITableViewCell.self, forCellReuseIdentifier: "Zeile")
        schatzListe.isScrollEnabled = true
        #endif
		raumListe.dataSource = self
    	raumListe.rowHeight = 20
    	raumListe.allowsMultipleSelection = false
    	#if os(macOS)
    	raumListe.headerView = nil
    	raumListe.delegate = self
    	raumListe.usesAutomaticRowHeights
    	raumListe.isEnabled = true
    	raumListe.selectionHighlightStyle = .regular
    	let col2 = NSTableColumn(identifier: NSUserInterfaceItemIdentifier(rawValue: "Zeile") )
		col2.width = 180
		raumListe.addTableColumn(col2)
    	//raumlistenScroll.is
    	#elseif os(iOS)
    	raumListe.allowsSelection = true
        raumListe.register(UITableViewCell.self, forCellReuseIdentifier: "Zeile")
        raumListe.isScrollEnabled = true
        #endif
        infoText.isEditable = false
        #if os(macOS)
        stärkeText.stringValue = "Stärke: 0"
        stärkeText.isEditable = false
        stärkeText.isSelectable = false
        stärkeText.drawsBackground = false
        stärkeText.isBordered = false
        handwaffenText.stringValue = "Handwaffe: keine"
        handwaffenText.isEditable = false
        handwaffenText.drawsBackground = false
        handwaffenText.isBordered = false
        kampfMonster.stringValue = "Sie kämpfen mit Monster:"
        kampfMonster.isEditable = false
        kampfMonster.drawsBackground = false
        kampfMonster.isBordered = false
        schlüsselAnzahl.stringValue = "Schlüssel: 0"
        schlüsselAnzahl.isEditable = false
        schlüsselAnzahl.drawsBackground = false
        schlüsselAnzahl.isBordered = false
        schutzwaffenText.stringValue = "Schutzwaffe: keine"
        schutzwaffenText.isEditable = false
        schutzwaffenText.drawsBackground = false
        schutzwaffenText.isBordered = false
        schatzText.stringValue = "Spielerschätze"
        schatzText.isEditable = false
        schatzText.alignment = .center
        schatzText.drawsBackground = false
        schatzText.isBordered = false
        raumText.stringValue = "Schätze im Raum"
        raumText.isEditable = false
        raumText.isSelectable = false
        raumText.alignment = .center
        raumText.drawsBackground = false
        raumText.isBordered = false
        gesamtGewicht.stringValue = "Gesamtgewicht: 0"
        gesamtGewicht.isEditable = false
        gesamtGewicht.drawsBackground = false
        gesamtGewicht.isBordered = false
        #elseif os(iOS)
        infoText.isScrollEnabled = true
        stärkeText.text = "Stärke: 0"
        handwaffenText.text = "Handwaffe: keine"
        kampfMonster.text = "Sie kämpfen mit Monster:"
        schlüsselAnzahl.text = "Schlüssel: 0"
        schutzwaffenText.text = "Schutzwaffe: keine"
        schatzText.text = "Spielerschätze"
        schatzText.textAlignment = .center
        raumText.text = "Schätze im Raum"
        raumText.textAlignment = .center
        gesamtGewicht.text = "Gesamtgewicht: 0"
        #endif
        kampfItems.isHidden = true

		#if os(macOS)
        handwaffeWeglegen.target = self
        schutzwaffeWeglegen.target = self
        spielregelAnzeigen.target = self
        mitnehmen.target = self
        dalassen.target = self
        aktivieren.target = self
        abbrechen.target = self
        weiterkämpfen.target = self
        schließenKnopf.target = self
		#elseif os(iOS)
        handwaffeWeglegen.setTitle("Handwaffe weglegen", for: .normal)
        handwaffeWeglegen.addTarget(self, action: #selector(Anzeige.KnopfGedrückt), for: .touchUpInside)
        schutzwaffeWeglegen.setTitle("Schutzwaffe weglegen", for: .normal)
        schutzwaffeWeglegen.addTarget(self, action: #selector(Anzeige.KnopfGedrückt), for: .touchUpInside)
        spielregelAnzeigen.setTitle("Spielregel", for: .normal)
        spielregelAnzeigen.addTarget(self, action: #selector(Anzeige.KnopfGedrückt), for: .touchUpInside)
        mitnehmen.setTitle("Mitnehmen", for: .normal)
        mitnehmen.addTarget(self, action: #selector(Anzeige.KnopfGedrückt), for: .touchUpInside)
        dalassen.setTitle("Dalassen", for: .normal)
        dalassen.addTarget(self, action: #selector(Anzeige.KnopfGedrückt), for: .touchUpInside)
        aktivieren.setTitle("Aktivieren", for: .normal)
        aktivieren.addTarget(self, action: #selector(Anzeige.KnopfGedrückt), for: .touchUpInside)
        abbrechen.setTitle("Kampf abbrechen", for: .normal)
        abbrechen.addTarget(self, action: #selector(Anzeige.KnopfGedrückt), for: .touchUpInside)
        weiterkämpfen.setTitle("Weiter kämpfen", for: .normal)
        weiterkämpfen.addTarget(self, action: #selector(Anzeige.KnopfGedrückt), for: .touchUpInside)
        schließenKnopf.setTitle("Schließen", for: .normal)
        schließenKnopf.addTarget(self, action: #selector(Anzeige.KnopfGedrückt), for: .touchUpInside)
        #endif
        handwaffeWeglegen.isEnabled = false
        schutzwaffeWeglegen.isEnabled = false
        spielregelAnzeigen.isEnabled = true
        mitnehmen.isEnabled = true
        dalassen.isEnabled = true
        aktivieren.isEnabled = true
        abbrechen.isEnabled = true
        weiterkämpfen.isEnabled = true
        
        hauptPanel.addSubview(mapping)
        #if os(macOS)
        hauptPanel.addSubview(schatzlistenScroll)
        #else
        hauptPanel.addSubview(schatzListe)
        #endif
        hauptPanel.addSubview(stärkeText)
        hauptPanel.addSubview(handwaffenText)
        #if os(macOS)
        hauptPanel.addSubview(infoTextScroll)
        #else
        hauptPanel.addSubview(infoText)
        #endif
        hauptPanel.addSubview(schlüsselAnzahl)
        hauptPanel.addSubview(schutzwaffenText)
        hauptPanel.addSubview (schatzText)
        hauptPanel.addSubview(kampfItems)
        hauptPanel.addSubview(raumItems)
        hauptPanel.addSubview(gesamtGewicht)
        hauptPanel.addSubview(handwaffeWeglegen)
        hauptPanel.addSubview(schutzwaffeWeglegen)
        hauptPanel.addSubview(aktivieren)
        hauptPanel.addSubview(dalassen)
        hauptPanel.addSubview(spielregelAnzeigen)
        #if os(macOS)
        raumItems.addSubview(raumlistenScroll)
        #else
        raumItems.addSubview(raumListe)
        #endif
        raumItems.addSubview(raumText)
        raumItems.addSubview(mitnehmen)
        kampfItems.addSubview(kampfMonster)
        kampfItems.addSubview(weiterkämpfen)
        kampfItems.addSubview(abbrechen)
        
		#if os(macOS)
        japadapadu.stringValue = "J a p a d a p a d u"
        japadapadu.drawsBackground = false
        japadapadu.isBordered = false
        gefunden.stringValue = "Der Ausgang ist gefunden"
        gefunden.drawsBackground = false
        gefunden.isBordered = false
		#elseif os(iOS)
        japadapadu.text = "J a p a d a p a d u"
        gefunden.text = "Der Ausgang ist gefunden"
        #endif
        endeMeldung.isHidden = true
        endeMeldung.addSubview(japadapadu)
        endeMeldung.addSubview(gefunden)
        spielregel.isHidden = true
        spielregel.addSubview(spielregelAnzeige)
        spielregel.addSubview(schließenKnopf)
        
        hauptPanel.isHidden = false
        addSubview(hauptPanel)
        addSubview(endeMeldung)
        addSubview(spielregel)
    }
    
    /**
    Wird beim Drücken der Knöpfe aufgerufen
    */
    @objc func KnopfGedrückt(sender:Button)
    {
        switch sender
        {
        case handwaffeWeglegen:
            kontroller.KommandoEintragen(kdo: Kommandos.handwaffeWeglegen, param: 0)
        case schutzwaffeWeglegen:
            kontroller.KommandoEintragen(kdo: Kommandos.schutzwaffeWeglegen, param: 0)
        case spielregelAnzeigen:
            spielregel.isHidden = false
            hauptPanel.isHidden = true
            spielregelAnzeige.SpielregelAnzeigen()
        case mitnehmen:
            #if os(macOS)
            kontroller.KommandoEintragen(kdo: Kommandos.schatzMitnehmen, param: raumListe.selectedRow)
            #elseif os(iOS)
            if let index = raumListe.indexPathForSelectedRow
            {
                kontroller.KommandoEintragen(kdo: Kommandos.schatzMitnehmen, param: index.row)
            }
            #endif
        case dalassen:
            #if os(macOS)
            kontroller.KommandoEintragen(kdo: Kommandos.schatzDalassen, param: schatzListe.selectedRow)
            #elseif os(iOS)
            if let index = schatzListe.indexPathForSelectedRow
            {
                kontroller.KommandoEintragen(kdo: Kommandos.schatzDalassen, param: index.row)
            }
            #endif
        case aktivieren:
            #if os(macOS)
            kontroller.KommandoEintragen(kdo: Kommandos.schatzAktivieren, param: schatzListe.selectedRow)
            #elseif os(iOS)
            if let index = schatzListe.indexPathForSelectedRow
            {
                kontroller.KommandoEintragen(kdo: Kommandos.schatzAktivieren, param: index.row)
            }
            #endif
        case weiterkämpfen:
            kontroller.KommandoEintragen(kdo: Kommandos.weiterKämpfen, param: 0)
        case abbrechen:
            kontroller.KommandoEintragen(kdo: Kommandos.kampfAbbrechen, param: 0)
        case schließenKnopf:
            spielregel.isHidden = true
            hauptPanel.isHidden = false
        default:
            break
        }
    }
  
    /**
	Setzt den Namen der Handwaffe
	- parameters:
		- name: Name der Handwaffe
     */
    func HandwaffenNameSetzen(name: String)
    {
        #if os(macOS)
        handwaffenText.stringValue = "Handwaffe: " + name
        #elseif os(iOS)
        handwaffenText.text = "Handwaffe: " + name
        #endif
        handwaffeWeglegen.isEnabled = name != "keine"
    }

    /**
	Setzt den Namen der Schutzwaffe
	- parameters:
		- name Name der Schutzwaffe
     */
    func SchutzwaffenNameSetzen(name: String)
    {
        #if os(macOS)
        schutzwaffenText.stringValue = "Schutzwaffe: " + name
        #elseif os(iOS)
        schutzwaffenText.text = "Schutzwaffe: " + name
        #endif
        schutzwaffeWeglegen.isEnabled = name != "keine"
    }

    /**
    Setzt die Stärkenanzeige
	- parameters:
		- st: Wert der Stärke
    */
    func StärkeSetzen(st: Int)
    {
        #if os(macOS)
        stärkeText.stringValue = "Stärke: \(st)"
        #elseif os(iOS)
        stärkeText.text = "Stärke: \(st)"
        #endif
    }

    /**
    Fügt eine neue Botschaft an die Anzeige an.
	- parameters:
		- text Botschaft
    */
    func NeueBotschaftEintragen(text: String)
    {
        #if os(macOS)
        infoText.string = text + "\n" + infoText.string
        #elseif os(iOS)
        infoText.text = text + "\n" + infoText.text
        #endif
    }

    /**
    Trägt den Monsternamen ein und blendet die nötigen Anzeigen für den Monsterkampf ein / aus.
	- parameters:
		- text der Monstername
    */
    func RundeVerlorenSetzen(text: String)
    {
        raumItems.isHidden = true
        mapping.isHidden = true
        kampfItems.isHidden = false
        #if os(macOS)
        kampfMonster.stringValue = "Sie kämpfen mit Monster: " + text
        #elseif os(iOS)
        kampfMonster.text = "Sie kämpfen mit Monster: " + text
        #endif
        dalassen.isEnabled = false
    }

    /**
    Beendet den Monsterkampf und blendet die nötigen Anzeigen ein / aus.
    */
    func KampfZuEndeSetzen()
    {
        kampfItems.isHidden = true
        mapping.isHidden = false
        dalassen.isEnabled = true
    }

	/**
	Zeigt dei Raumdaten an
    */
    private func RaumDatenAnzeigen()
    {
        raumItems.isHidden = false
    }

    /**
    Setzt die Anzahl der Schlüssel
	- parameters:
		- anz: Anzahl
    */
    func SchlüsselAnzahlSetzen(anz: Int)
    {
        #if os(macOS)
        schlüsselAnzahl.stringValue = "Schlüssel: \(anz)"
        #elseif os(iOS)
        schlüsselAnzahl.text = "Schlüssel: \(anz)"
        #endif
    }
    
    /**
    Setzt die Liste der Spielerschätze
	- parameters:
		- texte: Namen der Schätze
    */
    func SpielerSchätzeSetzen(texte: [String])
    {
        schatzlistenDaten = texte
        schatzListe.reloadData()
        dalassen.isEnabled = texte.count > 0
        aktivieren.isEnabled = texte.count > 0
    }
    
    /**
    Setzt die Liste der Raumschätze
	- parameters:
		- texte: Namen der Schätze
    */
    func RaumSchätzeSetzen(texte: [String])
    {
        raumlistenDaten = texte
        raumListe.reloadData()
        RaumDatenAnzeigen();
        mitnehmen.isEnabled = texte.count > 0
    }

    /**
    Setzt das Gesamtgewicht der Schätze
	- parameters:
		- gewicht: Gesamtgewicht
    */
    func GesamtgewichtSetzen(gewicht: Double)
    {
        #if os(macOS)
        gesamtGewicht.stringValue = "Gesamtgewicht: \(gewicht)"
        #elseif os(iOS)
        gesamtGewicht.text = "Gesamtgewicht: \(gewicht)"
        #endif
    }

    /**
    Löscht die Daten der Landkartenanzeige
    */
    func LandkartendatenLöschen()
    {
        mapping.LandkartendatenLöschen()
    }
    
    /**
    Trägt neue Rauminformation in die Landkartenanzeige ein.
	- parameters:
		- r: der einzutragende Raum
    */
    func RaumEintragen(r: LandkarteneintragRaum)
    {
        mapping.RaumEintragen(r: r)
    }
    
    /**
    Trägt neue Weginformation in die Landkartenanzeige ein.
	- parameters:
		- w: der einzutragende Weg
    */
    func WegEintragen(w: LandkarteneintragWeg)
    {
        mapping.WegEintragen(w: w)
    }
    
    /**
    Veranlasst das Neuzeichnen der Landkartenanzeige
    */
    func LandkarteNeuZeichnen()
    {
        #if os(macOS)
        mapping.needsDisplay = true
        #elseif os(iOS)
        mapping.setNeedsDisplay()
        #endif
    }
    
    /**
    Zeigt die Meldung über das erfolgreiche Spielende an.
    */
    func EndemeldungAnzeigen()
    {
        hauptPanel.isHidden = true
        endeMeldung.isHidden = false
    }
}



/**
Der eine Zufallsgenerator

- author: Albert Wiedemann
- version: 1.0
*/
class Zufall
{
	/** Das einzige Objekt der Klasse Zufall */
    static let zufall = Zufall()
	
    /**
    Initialisiert die verwendete Instanz des System-Zufallsgenerators
    */
    init()
    {
        srandom(0)
    }
    
    /**
    Typische Kommazufallszahl
    - returns: Eine Zufallszahl aus dem Intervall [0; 1[
    */
    func ZufallKomma() -> Float
    {
        return Float(arc4random()) / Float(UINT32_MAX)
    }

    /**
    Typische ganze Zufallszahl
    - returns: Eine ganze Zufallszahl aus dem Intervall [0; top[
    */
    func GanzzahlGeben(top: Int) -> Int
    {
        return Int(ZufallKomma() * Float(top))
    }
}


/**
Characterraster zur Kontrolle der zeichenbaren Wege

- author: Albert Wiedemann
- version: 1.0
*/
class Raster
{
    private let breite = breiteRaumRaster * 2 - 1
    private let höhe = höheRaumRaster * 2 - 1
    private let leerChar: Character = " "
    private let raumChar: Character = "%"
    private let wegChar: Character = "*"
    private var raster: [[Character]]

    /**
    Legt das Raster an und besetzt es mit Leerzeichen
    */
    init()
    {
        raster = [[Character]]()
        for i in 0 ..< breite
        {
            raster.append([Character]())
            for _ in 0 ..< höhe
            {
                raster[i].append(leerChar)
            }
        }
    }

    /**
	Meldet, ob der gegebene Rasterpunkt belegt ist
	- parameters:
		- x: x-Koordinate des Rasterpunkts
		- y: y-Koordinate des Rasterpunkts
	- returns: wahr, wenn der Punkt frei ist
    */
    func BelegtTesten(x: Int, y: Int) -> Bool
    {
        return raster [x] [y] != leerChar
    }
    
    /**
	Setzt den Rasterpunkt auf den gegebenen Wert
	- parameters:
		- x: x-Koordinate des Rasterpunkts
		- y: y-Koordinate des Rasterpunkts
		- was: der neue Wert
    */
    func Setzen(x: Int, y: Int, was: Character)
    {
        raster [x] [y] = was
    }
    
    /**
	Markiert den Rasterpunkt als Raum
	- parameters:
		- x: x-Koordinate des Raums
		- y: y-Koordinate des Raums
    */
    func RaumSetzen(x: Int, y: Int)
    {
        raster [x * 2] [y * 2] = raumChar
    }
    
    /**
	Meldet, ob an dem gegebene Rasterpunkt ein Raum gesetzt werden kann
	- parameters:
		- x: x-Koordinate des Raums
		- y: y-Koordinate des Raums
	- returns: wahr, wenn der Punkt frei ist
     */
    func RaumOkTesten(x: Int, y: Int) -> Bool
    {
        return (x >= 0) && (y >= 0) && (x < breiteRaumRaster) && (y < höheRaumRaster) && (raster [x * 2] [y * 2] == leerChar)
    }
    
    /**
	Meldet, zwischen den gegebenen Raumkoordinaten ein Weg gezeichnet werden kann
	- parameters:
		- x1p: x-Koordinate des ersten Rasterpunkts
		- y1p: y-Koordinate des ersten Rasterpunkts
		- x2p: x-Koordinate des zweiten Rasterpunkts
		- y2p: y-Koordinate des zweiten Rasterpunkts
	- returns: wahr, wenn der Weg frei ist
    */
    func WegTesten(x1p: Int, y1p: Int, x2p: Int, y2p: Int) -> Bool
    {
        var x1 = x1p * 2
        var x2 = x2p * 2
        var y1 = y1p * 2
        var y2 = y2p * 2
        while (x1 != x2) || (y1 != y2)
        {
            if x1 > x2
            {
                x1 -= 1
                x2 += 1
            }
            else if x1 < x2
            {
                x2 -= 1
                x1 += 1
            }
            if y1 > y2
            {
                y1 -= 1
                y2 += 1
            }
            else if y1 < y2
            {
                y2 -= 1
                y1 += 1
            }
            if (BelegtTesten(x: x1, y: y1) || BelegtTesten (x: x2, y: y2))
            {
                return false
            }
        }
        return true
    }
    
    /**
	Trägt den Weg zwischen den gegebenen Raumkoordinaten ein
	- parameters:
		- x1p: x-Koordinate des ersten Rasterpunkts
		- y1p: y-Koordinate des ersten Rasterpunkts
		- x2p: x-Koordinate des zweiten Rasterpunkts
		- y2p: y-Koordinate des zweiten Rasterpunkts
    */
    func WegZeichnen(x1p: Int, y1p: Int, x2p: Int, y2p: Int)
    {
        var x1 = x1p * 2
        var x2 = x2p * 2
        var y1 = y1p * 2
        var y2 = y2p * 2
        while (x1 != x2) || (y1 != y2)
        {
            if x1 > x2
            {
                x1 -= 1
                x2 += 1
            }
            else if x1 < x2
            {
                x2 -= 1
                x1 += 1
            };
            if y1 > y2
            {
                y1 -= 1
                y2 += 1
            }
            else if y1 < y2
            {
                y2 -= 1
                y1 += 1
            }
            Setzen(x: x1, y: y1, was: wegChar)
            Setzen(x: x2, y: y2, was: wegChar)
        }
    }
    
    /**
    Gibt das Raster zu Testzwecken ins Konsolfenster aus.
    */
    func Ausgeben()
    {
        for j in stride(from: höhe - 1, through: 0, by: -1)
        {
            for i in 0 ..< breite
            {
                print("\(raster [i] [j])", separator: "", terminator: "")
            }
            print()
        }
    }
}


/**
Verwaltet die Türen eines Raums.
- author: Albert Wiedemann
- version: 1.0
*/
class Türen
{
    /** Die erreichbaren Räume */
    var türen: [Raum]
    
    /**
    Der Konstruktor legt das Feld der Türen an.
    */
    init()
    {
        türen = [Raum]()
    }
    
    /**
	Prüft, ob es eine Tür zum Raum mit der angegebenen Nummer gibt
	- parameters:
		- nummer: die Raumnummer
	- returns: Raum oder null
    */
    func TürNachGeben(nummer: Int) -> Raum?
    {
        for r in türen
        {
            if r.nummer == nummer
            {
                return r
            }
        }
        return nil
    }

    /**
	Meldet die Anzahl der Türen
	- returns: Türenzahl
    */
    func AnzahlTürenGeben() -> Int
    {
        return türen.count
    }
    
    /**
	Meldet die Tür mit der angegebenen Nummer.
	- parameters:
		- nummer: die Türnummer
	- returns: Raum
    */
    func TürNummerGeben(nummer: Int) -> Raum
    {
        return türen[nummer]
    }
    
    /**
	Trägt eine neue Tür in die Türliste ein
	- parameters:
		- raum: der Raum, zu dem die Tür führen soll
    */
    func NeueTürEintragen(raum: Raum)
    {
        türen.append(raum);
    }
}

/**
Beschreibt einen Raum des Gewölbes.

- author: Albert Wiedemann
- version: 1.0
*/
class Raum
{
    /** Die Nummer des Raums */
    var nummer: Int
    /** die x-Koordinate des Raums in der Landkarte */
    let x: Int
    /** die v-Koordinate des Raums in der Landkarte */
    let y: Int
    /** Markierung, dass der Raum schon mal betreten wurde */
    var betreten: Bool
    /** Markierung, dass in dem Raum schon mal nach dem Ausgang gesucht wurde */
    var ausgangGesucht: Bool
    /** Die Türen zu den anderen Räumen */
    var türen: Türen
    /** Die Schätze im Raum */
    var schätze: [Schatz]
    /** Das Monster im Raum */
    var monster: Monster!

    /**
	der Konstruktor legt die Koordinaten und die Nummer fest.
	- parameters:
		- n: Raumnummer
		- x: x-Koordinate
		- y: y-Koordinate
    */
    init(n: Int, x: Int, y: Int)
    {
        nummer = n
        self.x = x
        self.y = y
        betreten = false
        ausgangGesucht = false
        türen = Türen()
        schätze = [Schatz]()
        monster = nil
    }

    /**
	Meldet die (erreichbaren) Raum mit der angegebenen Nummer.
	- parameters:
		- nummer: die gewünschte Raumnummer
	- returns: den zugehörigen Raum oder null
    */
    func TürNachGeben(nummer: Int) -> Raum?
    {
        return türen.TürNachGeben(nummer: nummer)
    }
    
    /**
	Trägt eine neue Tür in die Türliste ein
	- parameter:
		- raum: der Raum, zu dem die Tür führen soll
     */
    func NeueTürEintragen(nach: Raum)
    {
        türen.NeueTürEintragen(raum: nach);
    }

    /**
	Meldet, ob der Raum Schätze hat
	- reuturns: wahr, wenn mindestens ein Schatz vorhanden ist
     */
    func SchatzVorhanden() -> Bool
    {
        return !schätze.isEmpty
    }

    /**
	Lagert den Schatz in diesem Raum
	- parameters:
		- schatz: der zu lagernde Schatz
     */
    func SchatzLagern(schatz: Schatz)
    {
        schätze.append(schatz)
    }
    
    /**
	Gibt einen zufälligen Raum aus dem Gewölbe zurück
	- parameters:
		- zähler: Zähler, wie weit (noch) gesucht werden soll.
	- returns: Zufallsraum
     */
    func ZufallsRaumGeben(zähler: Int) -> Raum
    {
        if zähler == 0
        {
            return self
        }
        else
        {
            return türen.TürNummerGeben(nummer: Zufall.zufall.GanzzahlGeben(top: türen.AnzahlTürenGeben()))
        }
    }
    
    /**
	Versteckt den Schatz irgendwo im Gewölbe
	- parameters:
		- schatz der zu versteckende Schatz
    */
    func SchatzVerstecken(akt: Schatz)
    {
        var ziel: Raum
        repeat
        {
            ziel = ZufallsRaumGeben(zähler: 50)
        }
        while (ziel === self)
        ziel.SchatzLagern(schatz: akt)
    }
    
    /**
	Trägt das Betreten des Raumes und gegebenenfalls die stattgefundene Ausgangssuche ein
	- parameters:
		- ausgangOk: gibt an, ob auch nach dem Ausgang gesucht wird
     */
    func BetretenEintragen(ausgangOk: Bool)
    {
        betreten = true
        ausgangGesucht = ausgangGesucht || ausgangOk
    }
}


/**
Verwaltet das Gewölbe

- author: Albert Wiedemann
- version: 1.0
*/
class Gewölbe
{
    private var ausgang: Raum!
    private var markierungen: [Bool]
    private let spiel: Spiel
    private var räume: [Raum]

    /**
	Legt die Grundstrukturen an
     * @param spiel das Spiel, zu dem das Gewölbe gehört
     */
    init(spiel: Spiel)
    {
        ausgang = nil
        markierungen = [Bool](repeating: false, count: anzahlRäume)
        räume = [Raum]()
        self.spiel = spiel
    }

    /**
	Untersucht den Raum und seine Nachbarn
	- parameters:
		- akt: der aktuelle Raum
		- nummer: die Nummer des gesuchten Raums
	- returns: den gefundenen Raum oder null
    */
    func RaumLokalSuchen(akt: Raum, nummer: Int) -> Raum?
    {
        if !markierungen [akt.nummer]
        {
            markierungen [akt.nummer] = true
            if akt.nummer == nummer
            {
                return akt
            }
            else
            {
                for wohin in akt.türen.türen
                {
                    let hi = RaumLokalSuchen(akt: wohin, nummer: nummer)
                    if hi != nil
                    {
                        return hi
                    }
                }
                return nil
            }
        }
        return nil
    }

    /**
	Sucht den Raum mit der angegebenen Nummer
	- parameters:
		- nummer: die Nummer des gesuchten Raums
	- returns: den gefundenen Raum oder null
    */
    func RaumSuchen(nummer: Int) -> Raum
    {
        for i in 0 ..< anzahlRäume
        {
            markierungen [i] = false
        }
        if let r = RaumLokalSuchen(akt: ausgang, nummer: nummer)
        {
            return r
        }
        else
        {
        	fatalError("Raum \(nummer) konnte nicht gefunden werden in Gewölbe.RaumSuchen")
        }
    }

    /**
	Trägt die Daten für die Bildschirmdarstellung der Landkartenanzeige ein.
	- parameters:
		- alle: wenn wahr werden alle Räume gzeichnet, sonst nur die bekannten
		- markierung: der Raum, der markiert werden soll (i.d.R. der Raum des Spielers)
    */
    func AnzeigeZeichnen(alle: Bool, markierung: Raum)
    {
        var mode: Int
        spiel.anzeige.LandkartendatenLöschen();
        for akt in räume
        {
            if alle || akt.betreten
            {
                if akt === markierung
                {
                    mode = markiert
                }
                else if !alle && akt.betreten
                {
                    mode = besucht
                    if akt.ausgangGesucht
                    {
                        mode = ausgangGesucht
                    }
                }
                else
                {
                    mode = nichtBesucht
                }
                spiel.anzeige.RaumEintragen(r: LandkarteneintragRaum(nummer: akt.nummer, x: akt.x, y: akt.y, mode: mode))
                for wohin in akt.türen.türen
                {
                    spiel.anzeige.WegEintragen(w: LandkarteneintragWeg(x1: akt.x, y1: akt.y, x2: wohin.x, y2: wohin.y))
                    if !alle && !wohin.betreten
                    {
                        spiel.anzeige.RaumEintragen(r: LandkarteneintragRaum(nummer: wohin.nummer, x: wohin.x, y: wohin.y, mode: nichtBesucht))
                    }
                }
            }
        }
        spiel.anzeige.LandkarteNeuZeichnen()
    }

    /**
	Trägt einen neuen Raum in das Raster ein, falls der Platz frei ist und Weg vom aktuellen Raum dorthin möglich ist.
	- parameters:
		- von: der aktuelle Raum
		- x1: die x-Koordinaten des gewünschten Zielraums
		- y1: die y-Koordinaten des gewünschten Zielraums
		- nummer: die Nummer des gewünschten Raums
		- r: das Testraster für die Landspiel.anzeigendarstellung
    */
    func RaumEintragen(von: Raum, x1: Int, y1: Int, nummer: Int, r: Raster) -> Bool
    {
        if r.RaumOkTesten(x: x1, y: y1) && r.WegTesten(x1p: x1, y1p: y1, x2p: von.x, y2p: von.y)
        {
            r.RaumSetzen(x: x1, y: y1)
            r.WegZeichnen(x1p: x1, y1p: y1, x2p: von.x, y2p: von.y)
            let nach = Raum(n: nummer, x: x1, y: y1)
            nach.NeueTürEintragen(nach: von)
            von.NeueTürEintragen(nach: nach)
            räume.append(nach)
            return true
        }
        return false
    }
    
    /**
	Einen neuen Raum an den bestehenden Wegbaum anfügen
	- parameters:
		- nummer: die Nummer des gewünschten Raums
		- r: das Testraster für die Landspiel.anzeigendarstellung
     */
    func RaumAnfügen(nummer: Int, r: Raster)
    {
        var zähler = 0
        for _ in 1 ... anzahlRäume
        {
            repeat
            {
                let von = räume[Zufall.zufall.GanzzahlGeben(top: nummer)]
                var radius: Int
                switch Zufall.zufall.GanzzahlGeben(top: 10)
                {
                    case 8, 9:
                        radius = 3
                    case 5, 6, 7:
                        radius = 2
                    default:
                    	radius = 1
                }
                while radius > 0
                {
                    switch radius
                    {
                        case 1:
                            for x1 in von.x - 1 ... von.x + 1
                            {
                                for y1 in von.y - 1 ... von.y + 1
                                {
                                    if (x1 != von.x) || (y1 != von.y)
                                    {
                                        if (RaumEintragen(von: von, x1: x1, y1: y1, nummer: nummer, r: r))
                                        {
                                            return
                                        }
                                    }
                                }
                            }
                        case 2:
                            for x1 in von.x - 2 ... von.x + 2
                            {
                                if (RaumEintragen(von: von, x1: x1, y1: von.y - 2, nummer: nummer, r: r))
                                {
                                    return
                                }
                            }
                            for x1 in von.x - 2 ... von.x + 2
                            {
                                if (RaumEintragen(von: von, x1: x1, y1: von.y + 2, nummer: nummer, r: r))
                                {
                                    return
                                }
                            }
                            for y1 in von.y - 1 ... von.y + 1
                            {
                                if (RaumEintragen(von: von, x1: von.x - 2, y1: y1, nummer: nummer, r: r))
                                {
                                    return
                                }
                            }
                            for y1 in von.y - 1 ... von.y + 1
                            {
                                if (RaumEintragen(von: von, x1: von.x + 2, y1: y1, nummer: nummer, r: r))
                                {
                                    return
                                }
                            }
                        case 3:
                            for x1 in von.x - 3 ... von.x + 3
                            {
                                if (RaumEintragen(von: von, x1: x1, y1: von.y - 3, nummer: nummer, r: r))
                                {
                                    return
                                }
                            }
                            for x1 in von.x - 3 ... von.x + 3
                            {
                                if (RaumEintragen(von: von, x1: x1, y1: von.y + 3, nummer: nummer, r: r))
                                {
                                    return
                                }
                            }
                            for y1 in von.y - 2 ... von.y + 2
                            {
                                if (RaumEintragen(von: von, x1: von.x - 3, y1: y1, nummer: nummer, r: r))
                                {
                                    return
                                }
                            }
                            for y1 in von.y - 2 ... von.y + 2
                            {
                                if (RaumEintragen(von: von, x1: von.x + 3, y1: y1, nummer: nummer, r: r))
                                {
                                    return
                                }
                            }
                        default:
                            break
                    }
                    radius -= 1
                }
                zähler += 1
            }
            while zähler < anzahlRäume * 2
        }
        fatalError("Gewölbe.Raumanfügen: keinen Platz für Raum gefunden")
    }
    
    /**
	Erzeugt einen Weg zwischen zwei zufälligen Räumen
	- parameters:
		- r: das Testraster für die Landspiel.anzeigendarstellung
     */
    func ZufallsWegErzeugen(r: Raster)
    {
        let maxRadius2 = breiteRaumRaster * breiteRaumRaster + höheRaumRaster  * höheRaumRaster
        let von = räume[Zufall.zufall.GanzzahlGeben(top: anzahlRäume)]
        var radius = 0
        repeat
        {
            radius += 1
            for nummer in 0 ..< anzahlRäume
            {
                let nach = räume[nummer]
                var dx = nach.x - von.x
                var dy = nach.y - von.y
                dx = dx * dx
                dy = dy * dy
                if ((radius - 1) * (radius - 1) < dx + dy) && (radius * radius >= dx + dy) &&
                    r.WegTesten(x1p: von.x, y1p: von.y, x2p: nach.x, y2p: nach.y)
                {
                    r.WegZeichnen(x1p: von.x, y1p: von.y, x2p: nach.x, y2p: nach.y)
                    nach.NeueTürEintragen(nach: von)
                    von.NeueTürEintragen(nach: nach)
                    return
                }
            }
        }
        while radius * radius < maxRadius2
    }

    /**
    Tauscht die Raumnummer zufällig durch
    */
    func NummernTauschen()
    {
        for _ in 1 ... anzahlRäume * 2
        {
            let von = räume[Zufall.zufall.GanzzahlGeben(top: anzahlRäume)]
            let nach = räume[Zufall.zufall.GanzzahlGeben(top: anzahlRäume)]
            let hilf = nach.nummer
            nach.nummer = von.nummer
            von.nummer = hilf
        }
    }

    /**
	Erzeugt alle Schätze im Gewölbe
    */
    func SchätzeGenerieren()
    {
        var akt: Raum
        for _ in 1 ... 3
        {
            repeat
            {
                akt = räume[Zufall.zufall.GanzzahlGeben(top: anzahlRäume)]
            }
            while !akt.schätze.isEmpty
            akt.schätze.append(Schlüssel(sp: spiel))
        }
        for _ in 1 ... anzahlRäume * 3 / 4
        {
            repeat
            {
                akt = räume[Zufall.zufall.GanzzahlGeben(top: anzahlRäume)]
            }
            while !akt.schätze.isEmpty
            akt.schätze.append(Schatz.ZufälligGenerieren(spiel: spiel))
        }
    }
    
    /**
	Erzeugt alle Monster im Gewölbe
    */
    func MonsterGenerieren()
    {
        var akt: Raum
        for _ in 1 ... anzahlRäume * 2 / 3
        {
            repeat
            {
                akt = räume[Zufall.zufall.GanzzahlGeben(top: anzahlRäume)]
            }
            while akt.monster != nil
            akt.monster = Monster.ZufälligGenerieren(spiel: spiel);
        }
        for akt in räume
        {
            if akt.monster == nil
            {
                akt.monster = LeeresMonster(sp: spiel)
            }
        }
    }
    
    /**
	Erzeugt eine neues Gewölbe
    */
    func Generieren()
    {
        let r = Raster()
        ausgang = Raum(n: 0, x: Zufall.zufall.GanzzahlGeben(top: breiteRaumRaster), y: Zufall.zufall.GanzzahlGeben(top: höheRaumRaster))
        räume.append(ausgang)
        r.RaumSetzen(x: ausgang.x, y: ausgang.y)
        for i in 1 ..< anzahlRäume
        {
            RaumAnfügen(nummer: i, r: r)
        }
        for _ in 1 ... anzahlRäume
        {
            ZufallsWegErzeugen(r: r)
        }
        NummernTauschen()
        SchätzeGenerieren()
        MonsterGenerieren()
    }

    /**
	Gibt einen zufällig gewählten Startraum zurück
	- returns: Startraum
     */
    func StartRaum() -> Raum
    {
        return RaumSuchen(nummer: Zufall.zufall.GanzzahlGeben(top: anzahlRäume))
    }
    
    /**
	Testet, ob ein Raum der Ausgang ist
	- parameters:
		- test: der zu testende Raum
	- returns: wahr, wenn der Raum der Ausgang ist.
     */
    func AusgangTesten(test: Raum) -> Bool
    {
        return ausgang === test
    }
}


/**
Oberklasse aller Objekte des Spiels.

- author: Albert Wiedemann
- version: 1.0
*/
class SpielObjekt
{
    /** Spiel, zu dem das Objekt gehört */
    let spiel: Spiel
    
    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
    */
    init(sp: Spiel)
    {
        spiel = sp
    }
    
    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    func NameGeben() -> String
    {
        return "Allgemeines Spielobjekt"
    }
}


/**
Oberklasse aller Monster

- author: Albert Wiedemann
- version: 1.0
*/
class Monster: SpielObjekt
{
    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
    */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
    }
    
    /**
	Generiert ein zufälliges Monster
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
	- returns: das erzeugt Monster
    */
    static func ZufälligGenerieren(spiel: Spiel) -> Monster
    {
        switch Zufall.zufall.GanzzahlGeben(top: 6)
        {
            case 0, 1:
                return Dieb(sp: spiel)
            case 2, 3:
                return Schlange(sp: spiel)
            case 4:
                return Drache(sp: spiel)
            case 5:
                return Falltür(sp: spiel)
            default:
                fatalError("Unbekannter Monsterwert in Monster.ZufälligGenerieren")
        }
    }
    
    /**
	Kampf des Spielers mit dem Monster
	- parameters:
		- sp: der angegriffende Spieler
	- returns: wahr, wenn der Spieler den Kampf gewonnen hat.
    */
    func KampfDurchführen(sp: Spieler) -> Bool
    {
        return true
    }
    
    /**
	Einleitung des Kampfs.
	
	In der Regel die Meldung, welches Monster angreift.
	- parameters:
		- sp: der angegriffene Spieler
    */
    func KampfBeginnen(sp: Spieler)
    {
        spiel.BotschaftSenden(botschaft: "Ein Monster fällt Sie an: " + NameGeben ())
    }
    
    /**
	Kampfabbruch durch Flucht des Spielers.
	
	Standardmäßig keine Aktion
	- parameters:
		- sp: der angegriffene Spieler
    */
    func KampfAbbrechen(sp: Spieler)
    {
    }
}

/**
Das Monster Dieb.

Der Dieb klaut dem Spieler einen Schatz, wen dieser keine Handwaffe hat.
- author: Albert Wiedemann
- version: 1.0
*/
class Dieb: Monster
{
    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
    */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
    }
    
    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    override func NameGeben() -> String
    {
        return "Dieb"
    }
    
    /**
	Einleitung des Kampfs.
	
	Der Dieb meldet sich nur, wenn er klauen kann.
	- parameters:
		- sp: der angegriffene Spieler
    */
    override func KampfBeginnen(sp: Spieler)
    {
        if !sp.schätze.isEmpty && (sp.handWaffe == nil)
        {
            super.KampfBeginnen(sp: sp)
        }
    }
    
    /**
	Kampf des Spielers mit dem Monster
	- parameters:
		- sp: der angegriffende Spieler
	- returns: wahr, wenn der Spieler den Kampf gewonnen hat.
    */
    override func KampfDurchführen(sp: Spieler) -> Bool
    {
        if !sp.schätze.isEmpty && (sp.handWaffe == nil)
        {
            while !sp.schätze.isEmpty && (Zufall.zufall.GanzzahlGeben(top: 2) == 0)
            {
                sp.wo.SchatzLagern(schatz: sp.schätze.remove(at: Zufall.zufall.GanzzahlGeben(top: sp.schätze.count)))
            }
            return false
        }
        else
        {
            return true
        }
    }
}

/**
Das Monster Drache.

Der Drache klaut dem Spieler Schätze, wenn er keine genügend starke Handwaffe hat.
- author: Albert Wiedemann
- version: 1.0
*/
class Drache: Monster
{
    /** Die Zahl der runden im Monsterkampf */
    private var runden: Int

    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
    */
    override init(sp: Spiel)
    {
        runden = 0
        super.init(sp: sp)
    }

    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    override func NameGeben() -> String
    {
        return "Drache"
    }
    
    /**
	Einleitung des Kampfs.
	- parameters:
		- sp: der angegriffene Spieler
    */
    override func KampfBeginnen(sp: Spieler)
    {
        super.KampfBeginnen(sp: sp)
        runden = 0
    }
    
    /**
	Kampf des Spielers mit dem Monster
	- parameters:
		- sp: der angegriffende Spieler
	- returns: wahr, wenn der Spieler den Kampf gewonnen hat.
    */
    override func KampfDurchführen(sp: Spieler) -> Bool
    {
        runden += 1
        if (sp.handWaffe != nil) && (sp.stärke + sp.handWaffe!.WaffenstärkeGeben() + Zufall.zufall.GanzzahlGeben(top: 15) >= 35)
        {
            sp.handWaffe!.AnwendungEintragen()
            spiel.BotschaftSenden(botschaft: "Diesen Drachen haben Sie besiegt")
            sp.StärkeÄndern(wert: -2)
            if (sp.schutz == nil) || (sp.schutz!.AbwehrstärkeGeben() <= Zufall.zufall.GanzzahlGeben(top: 15))
            {
                sp.StärkeÄndern(wert: -2)
            }
            else
            {
                sp.schutz!.AnwendungEintragen()
            }
            sp.wo.monster = LeeresMonster(sp: spiel)
            return true
        }
        else
        {
            sp.StärkeÄndern(wert: -1);
            if (sp.schutz == nil) || (sp.schutz!.AbwehrstärkeGeben() <= Zufall.zufall.GanzzahlGeben(top: 15))
            {
                sp.StärkeÄndern(wert: -1)
            }
            return false
        }
    }
    
    /**
	Kampfabbruch durch Flucht des Spielers.
	- parameters:
		- sp: der angegriffene Spieler
    */
    override func KampfAbbrechen(sp: Spieler)
    {
        if !sp.schätze.isEmpty
        {
            sp.wo.SchatzLagern(schatz: sp.schätze.remove(at: Zufall.zufall.GanzzahlGeben(top: sp.schätze.count)))
            if (runden*2 >= sp.stärke) && !sp.schätze.isEmpty
            {
                sp.wo.SchatzLagern(schatz: sp.schätze.remove(at: Zufall.zufall.GanzzahlGeben(top: sp.schätze.count)))
            }
        }
    }
}

/**
Das Monster Falltür.

Der Spieler wird irgendwohin versetzt
- author: Albert Wiedemann
- version: 1.0
*/
class Falltür: Monster
{
    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
    */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
    }

    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    override func NameGeben() -> String
    {
        return "Falltür"
    }
    
    /**
	Einleitung des Kampfs.
	- parameters:
		- sp: der angegriffene Spieler
    */
    override func KampfBeginnen(sp: Spieler)
    {
        Falle(woher: sp.wo, s: spiel)
    }
    
    /**
	Kampf des Spielers mit dem Monster
	- parameters:
		- sp: der angegriffende Spieler
	- returns: wahr, wenn der Spieler den Kampf gewonnen hat.
    */
    override func KampfDurchführen(sp: Spieler) -> Bool
    {
        return true
    }

    class Falle: Türen
    {
        /** die ursprünglichen Türen */
        private var türenOrig: Türen
        /** der betroffene Raum */
        private var woher: Raum
        /** das aktuelle Spiel*/
        private let spiel: Spiel
        
        /**
        Merkt sich die Situation.
        */
        init(woher: Raum, s: Spiel)
        {
            self.woher = woher
            spiel = s
            türenOrig = woher.türen
            super.init()
            woher.türen = self
        }
        
        /**
		Meldet, ob das Ziel der eigene Raum oder ein Nachbarraum ist
		- parameters:
			- ziel: eigener Raum
		- returns: wahr, wenn das Ziel zu nahe ist
         */
        private func ZielZunahe(ziel: Raum) -> Bool
        {
            if ziel === woher
            {
                return true
            }
            for wohin in türenOrig.türen
            {
                if wohin === ziel
                {
                    return true
                }
            }
            return false
        }
        
        /**
        Prüft, ob es eine Tür zum Raum mit der angegebenen Nummer gibt
        - parameters:
            - nummer: die Raumnummer
        - returns: Raum oder null
        */
        override func TürNachGeben(nummer: Int) -> Raum?
        {
            var hilf = türenOrig.TürNachGeben(nummer: nummer)
            if !(woher.monster is Falltür)
            {
                woher.türen = türenOrig
            }
            else if hilf != nil
            {
                spiel.BotschaftSenden(botschaft: "Schönen Gruß von der Falltür.");
                woher.türen = türenOrig
                repeat
                {
                    hilf = woher.ZufallsRaumGeben(zähler: 50)
                } while (ZielZunahe(ziel: hilf!))
            }
            return hilf
        }
        
        /**
        Meldet die Anzahl der Türen
        - returns: Türenzahl
        */
        override func AnzahlTürenGeben() -> Int
        {
            return türenOrig.AnzahlTürenGeben()
        }
    
        /**
        Meldet die Tür mit der angegebenen Nummer.
        - parameters:
            - nummer: die Türnummer
        - returns: Raum
        */
        override func TürNummerGeben(nummer: Int) -> Raum
        {
            return türenOrig.TürNummerGeben(nummer: nummer)
        }
    }
}

/**
Das Monster Schlange.

Die Schlange klaut dem Spieler Stärke, wenn er keine genügend starke Handwaffe hat.
- author: Albert Wiedemann
- version: 1.0
*/
class Schlange: Monster
{
    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
    */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
    }

    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    override func NameGeben() -> String
    {
        return "Schlange"
    }
    
    /**
	Einleitung des Kampfs.
	- parameters:
		- sp: der angegriffene Spieler
    */
    override func KampfBeginnen(sp: Spieler)
    {
        super.KampfBeginnen(sp: sp);
        if sp.schutz == nil
        {
            sp.StärkeÄndern(wert: -sp.stärke / 2)
        }
        else if sp.schutz!.AbwehrstärkeGeben() <= 5 + Zufall.zufall.GanzzahlGeben(top: 7)
        {
            sp.StärkeÄndern(wert: -sp.stärke / 4)
        }
        else
        {
            sp.schutz!.AnwendungEintragen();
        }
    }
    
    /**
	Kampf des Spielers mit dem Monster
	- parameters:
		- sp: der angegriffende Spieler
	- returns: wahr, wenn der Spieler den Kampf gewonnen hat.
    */
    override func KampfDurchführen(sp: Spieler) -> Bool
    {
        if (sp.handWaffe != nil) && (sp.stärke + sp.handWaffe!.WaffenstärkeGeben() + Zufall.zufall.GanzzahlGeben(top: 5) >= 15)
        {
            sp.handWaffe!.AnwendungEintragen()
            spiel.BotschaftSenden(botschaft: "Diese Schlange ist aufgeräumt")
            sp.wo.monster = LeeresMonster(sp: spiel);
            return true
        }
        else
        {
            sp.StärkeÄndern(wert: -1);
            if (sp.schutz == nil) || (sp.schutz!.AbwehrstärkeGeben() <= 5 + Zufall.zufall.GanzzahlGeben(top: 7))
            {
                sp.StärkeÄndern(wert: -2)
            }
            return false
        }
    }

    /**
	Kampfabbruch durch Flucht des Spielers.
	- parameters:
		- sp: der angegriffene Spieler
    */
    override func KampfAbbrechen(sp: Spieler)
    {
        if !sp.schätze.isEmpty
        {
            sp.wo.SchatzLagern(schatz: sp.schätze.remove(at: Zufall.zufall.GanzzahlGeben(top: sp.schätze.count)))
            if !sp.schätze.isEmpty
            {
                sp.wo.SchatzLagern(schatz: sp.schätze.remove(at: Zufall.zufall.GanzzahlGeben(top: sp.schätze.count)))
            }
        }
    }
}

/**
Simuliert "keiner da."

- author: Albert Wiedemann
- version: 1.0
*/
class LeeresMonster: Monster
{
    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
    }

    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    override func NameGeben() -> String
    {
        return "Leeres Monster"
    }
    
    /**
	Einleitung des Kampfs.
	
	Keine Meldung.
	- parameters:
		- sp: der angegriffene Spieler
    */
    override func KampfBeginnen(sp: Spieler)
    {
    }
    
    /**
	Kampf des Spielers mit dem Monster
	- parameters:
		- sp: der angegriffende Spieler
	- returns: wahr, wenn der Spieler den Kampf gewonnen hat.
    */
    override func KampfDurchführen(sp: Spieler) -> Bool
    {
        return true
    }
}

/**
Oberklasse aller Schätze

- author: Albert Wiedemann
- version: 1.0
*/
class Schatz: SpielObjekt
{
    /** Das Gewicht des Schatzes */
    var gewicht: Double

    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        gewicht = 0.0
        super.init(sp: sp)
    }

    /**
     * Generiert einen zufälligen Schatz
     * @param sp das Spiel, in dem das Objekt verwendet wird.
     * @return der erzeugte Schatz
     */
    static func ZufälligGenerieren(spiel: Spiel) -> Schatz
    {
        switch (Zufall.zufall.GanzzahlGeben(top: 17))
        {
            case 0, 1:
                return Schwert(sp: spiel)
            case 2, 3, 4:
                return Dolch(sp: spiel)
            case 5:
                return Zweihänder(sp: spiel)
            case 6:
                return Rüstung(sp: spiel)
            case 7, 8:
                return Helm(sp: spiel)
            case 9, 10, 11:
                return Lederkappe(sp: spiel)
            case 12:
                return Feuerball(sp: spiel)
            case 13:
                return RuckSack(sp: spiel)
            case 14:
                return Karte(sp: spiel)
            case 15:
                return Heiltrank(sp: spiel)
            case 16:
                return Bogen(sp: spiel)
            default:
            	fatalError("Unzulässige Nummer in Schatz.ZufälligGenerieren")
        }
    }
    
    /**
	Meldet das Gewicht des Schatzes
	- returns: Gewicht
    */
    func GewichtGeben() -> Double
    {
        return gewicht
    }
}


/**
Hebt das Gewicht der Schätzte auf

- author: Albert Wiedemann
- version: 1.0
*/
class RuckSack: Schatz
{

    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
        gewicht = -100.0
    }

    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    override func NameGeben() -> String
    {
        return "Zauberrucksack"
    }
}

/**
Der Schlüssel zum Ausgang

- author: Albert Wiedemann
- version: 1.0
*/
class Schlüssel: Schatz
{
    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
        gewicht = 0.1
    }

    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    override func NameGeben() -> String
    {
        return "Schlüssel"
    }
}

/**
Oberklasse aller Sprüche

- author: Albert Wiedemann
- version: 1.0
*/
class Spruch: Schatz
{
    /** Ein Spruch hat ein Mentalgewicht, man muss ihn sich merken */
    var mentalgewicht: Int

    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        mentalgewicht = Int.max
        super.init(sp: sp)
    }

    /**
    Meldet das Mentalgewicht des Spruchs.
	- returns: Mentalgewicht
     */
    func MentalgewichtGeben() -> Int
    {
        return mentalgewicht
    }
    
    /**
	Wird aufgerufen, wenn der Spruch gefunden wurde.
	
	Die Oberklassenimplemetierung hat die leere Leistung
    */
    func GefundenMelden()
    {
        //nichts
    }

    /**
    Führt den Spruch aus
    */
    func Ausführen()
    {
    }
}

/**
Der Feuerball verjagt das Monster im Zielraum

- author: Albert Wiedemann
- version: 1.0
*/
class Feuerball: Spruch
{
    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
        gewicht = 3.0
        mentalgewicht = 3
    }

    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    override func NameGeben() -> String
    {
        return "Feuerball"
    }
    
    /**
    Führt den Spruch aus
    */
    override func Ausführen()
    {
        spiel.spieler.schätze.remove(at: spiel.spieler.schätze.firstIndex(where: {$0 === self})!)
        let ort = spiel.spieler.wo
        //spiel.spieler.wo.monster = LeeresMonster(sp: spiel)
        ort.monster = LeeresMonster(sp: spiel)
        //spiel.spieler.wo.SchatzVerstecken(akt: self)
        ort.SchatzVerstecken(akt: self)
        spiel.spieler.InventarListeEintragen()
    }
}

/**
Der Heiltrank gibt dem Spieler pro Anwendung einen Stärkepunkt zurück

- author: Albert Wiedemann
- version: 1.0
*/
class Heiltrank: Spruch
{
    /** Anzahl der noch möglichen Verwendungen */
    private var anzahl: Int

    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        anzahl = 0
        super.init(sp: sp)
        gewicht = 2.0
        mentalgewicht = 5
    }

    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    override func NameGeben() -> String
    {
        if (anzahl > 0)
        {
            return "Heiltrank (\(anzahl))"
        }
        else
        {
            return "Heiltrank"
        }
    }
    
    /**
    Setzt die Anzahl der Anwendungen
    */
   override func GefundenMelden()
    {
        anzahl = 5
    }
    
    /**
    Führt den Spruch aus
    */
    override func Ausführen()
    {
        spiel.spieler.StärkeÄndern(wert: 1)
        anzahl = anzahl - 1
        if anzahl == 0
        {
            spiel.spieler.schätze.remove(at: spiel.spieler.schätze.firstIndex(where: {$0 === self})!)
            //%%%%%spiel.spieler.wo.SchatzVerstecken(akt: self)
            let ort = spiel.spieler.wo
            ort.SchatzVerstecken(akt: self)
            spiel.spieler.InventarListeEintragen()
        }
    }
}

/**
Zeigt die gesamte Landkarte an.
 
- author: Albert Wiedemann
- version: 1.0
*/
class Karte: Spruch
{
    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
        gewicht = 1.0
        mentalgewicht = 5
    }

    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    override func NameGeben() -> String
    {
        return "Karte"
    }
    
    /**
    Führt den Spruch aus
    */
    override func Ausführen()
    {
        spiel.gewölbe.AnzeigeZeichnen(alle: true, markierung: spiel.spieler.wo);
    }
}

/**
Oberklasse aller Waffen

- author: Albert Wiedemann
- version: 1.0
*/
class Waffe: Schatz
{
    /** Zahl der Anwendungen */
    var anzahlAnwendungen: Int
    
    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        anzahlAnwendungen = 0
        super.init(sp: sp)
    }
    
    /**
    Trägt ein, dass eine Waffe verwendet wurde.
    */
    func AnwendungEintragen()
    {
        anzahlAnwendungen = anzahlAnwendungen + 1
    }
}

/**
Oberklasse aller Angriffswaffen

- author: Albert Wiedemann
- version: 1.0
*/
class AngriffsWaffe: Waffe
{

    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
    }

    /**
	Meldet die Stärke der Angriffskraft
	- returns: Angriffsstärke
     */
    func WaffenstärkeGeben() -> Int
    {
        return 0
    }
}

/**
Der Bogen hat eine mittlere Angrifsstärke

- author: Albert Wiedemann
- version: 1.0
*/
class Bogen: AngriffsWaffe
{

    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
        gewicht = 8.0
    }

    /**
	Meldet die Stärke der Angriffskraft
	- returns: Angriffsstärke
     */
    override func WaffenstärkeGeben() -> Int
    {
        return 10
    }
    
    /**
    Trägt ein, dass eine Waffe verwendet wurde.
    */
    override func AnwendungEintragen()
    {
        super.AnwendungEintragen()
        if anzahlAnwendungen >= 12
        {
            spiel.spieler.schätze.remove(at: spiel.spieler.schätze.firstIndex(where: {$0 === self})!)
            //%%%%%spiel.spieler.wo.SchatzVerstecken(akt: self)
            let ort = spiel.spieler.wo
            ort.SchatzVerstecken(akt: self)
            spiel.spieler.InventarListeEintragen()
            anzahlAnwendungen = 0
        }
    }
    
    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    override func NameGeben() -> String
    {
        return "Bogen"
    }
}

/**
Der Dolch hat eine geringe Angrifsstärke

- author: Albert Wiedemann
- version: 1.0
*/
class Dolch: AngriffsWaffe
{

    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
        gewicht = 2.0
    }

    /**
	Meldet die Stärke der Angriffskraft
	- returns: Angriffsstärke
     */
    override func WaffenstärkeGeben() -> Int
    {
        return 5
    }
    
    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    override func NameGeben() -> String
    {
        return "Dolch"
    }
}

/**
Das Schwert hat eine mittlere Angrifsstärke

- author: Albert Wiedemann
- version: 1.0
*/
class Schwert: AngriffsWaffe
{
    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
        gewicht = 10.0
    }

    /**
	Meldet die Stärke der Angriffskraft
	- returns: Angriffsstärke
     */
    override func WaffenstärkeGeben() -> Int
    {
        return 11
    }
    
    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    override func NameGeben() -> String
    {
        return "Schwert"
    }
}

/**
Der Zweihänder hat eine große Angrifsstärke

- author: Albert Wiedemann
- version: 1.0
*/
class Zweihänder: AngriffsWaffe
{
    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
        gewicht = 20.0
    }

    /**
	Meldet die Stärke der Angriffskraft
	- returns: Angriffsstärke
     */
    override func WaffenstärkeGeben() -> Int
    {
        return 16
    }
    
    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    override func NameGeben() -> String
    {
        return "Zweihänder"
    }
}

/**
Oberklasse aller Schutzwaffen

- author: Albert Wiedemann
- version: 1.0
*/
class SchutzWaffe: Waffe
{
    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
    }

    /**
	Meldet die Stärke der Abwehrkraft
	- returns: Abwehrstärke
     */
    func AbwehrstärkeGeben() -> Int
    {
        return 0
    }

}

/**
Der Helm hat eine mittlere Abwekrstärke

- author: Albert Wiedemann
- version: 1.0
*/
class Helm: SchutzWaffe
{
    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
        gewicht = 15.0
    }

    /**
	Meldet die Stärke der Abwehrkraft
	- returns: Abwehrstärke
     */
    override func AbwehrstärkeGeben() -> Int
    {
        return 9 - anzahlAnwendungen / 2
    }
    
    /**
    Trägt ein, dass eine Waffe verwendet wurde.
    */
    override func AnwendungEintragen()
    {
        super.AnwendungEintragen()
        if anzahlAnwendungen >= 18
        {
            spiel.spieler.schätze.remove(at: spiel.spieler.schätze.firstIndex(where: {$0 === self})!)
            //%%%%%spiel.spieler.wo.SchatzVerstecken(akt: self)
            let ort = spiel.spieler.wo
            ort.SchatzVerstecken(akt: self)
            spiel.spieler.InventarListeEintragen()
            anzahlAnwendungen = 0
        }
    }
    
    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    override func NameGeben() -> String
    {
        return "Helm"
    }
}

/**
Die Lederkappe hat eine geringe Abwekrstärke

- author: Albert Wiedemann
- version: 1.0
*/
class Lederkappe: SchutzWaffe
{
    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
        gewicht = 5.0
    }

    /**
	Meldet die Stärke der Abwehrkraft
	- returns: Abwehrstärke
     */
    override func AbwehrstärkeGeben() -> Int
    {
        return 3
    }
    
    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    override func NameGeben() -> String
    {
        return "Lederkappe"
    }
}

/**
Die Rüstung hat eine hohe Abwekrstärke

- author: Albert Wiedemann
- version: 1.0
*/
class Rüstung: SchutzWaffe
{
    /**
	Der Konstruktor besetzt das Spiel
	- parameters:
		- sp: das Spiel, in dem das Objekt verwendet wird.
     */
    override init(sp: Spiel)
    {
        super.init(sp: sp)
        gewicht = 50.0
    }

    /**
	Meldet die Stärke der Abwehrkraft
	- returns: Abwehrstärke
     */
    override func AbwehrstärkeGeben() -> Int
    {
        return 20
    }
    
    /**
	Meldet den Namen des Objekts zurück
	- returns: name
    */
    override func NameGeben() -> String
    {
        return "Rüstung"
    }
}

/**
Verwaltet das Spielerobjekt

- author: Albert Wiedemann
- version: 1.0
*/
class Spieler: SpielObjekt
{
    /** das maximal tragbare Geewicht */
    private let maxGewicht = 90.0
    /** die maximale Stärke */
    private let maxStärke = 20
    /** der Raum, in dem sich der Spieler zur Zeit befindet */
    var wo: Raum
    /** die Schätze des Spielers */
    var schätze: [Schatz]
    /** die Handwaffe */
    var handWaffe: AngriffsWaffe?
    /** die Schutzwaffe */
    var schutz: SchutzWaffe?
    /** die aktuelle Stärke */
    var stärke: Int
    /** die Erfahrung */
    private var erfahrung: Int

    /**
	Der Konstruktor setzt die Rahmendaten
	- parameters:
		- spiel: das Spiel, in dem der Spieler agiert
		- start: der Startraum des Spielers
     */
    init(spiel: Spiel, start: Raum)
    {
        wo = start
        wo.betreten = true
        schätze = [Schatz]()
        handWaffe = nil
        schutz = nil
        stärke = maxStärke
        erfahrung = 0
        super.init(sp: spiel)
    }

    /**
	Meldet den Namen des Objekts zurück
	- returns: name
     */
    override func NameGeben() -> String
    {
        return "Spieler"
    }

    /**
	Geht zum angegebenen Raum
	- parameters:
		- nummer: Nummer des Zielraums
	- returns: true, wenn der Raum erreichbar ist
     */
    func NachRaumGehen(nummer: Int) -> Bool
    {
        
        if let neu = wo.TürNachGeben(nummer: nummer)
        {
            wo = neu
            wo.BetretenEintragen(ausgangOk: AnzahlSchlüsselGeben() == 3)
            return true
        }
        else
        {
            return false
        }
    }

    /**
	Ermittelt die Anzahl der Schlüssel des Spielers
	- returns Anzahl der Schlüssel
     */
    func AnzahlSchlüsselGeben() -> Int
    {
        var anz = 0
        for akt in schätze
        {
            if akt is Schlüssel
            {
                anz += 1
            }
        }
        return anz
    }
    
    /**
	Prüft, ob alle Schätze getragen werden können und lässt gegebenenfalls welche da.
	
	Sonderschätze werden immer behalten, Hand- und Schutzwaffe wenn möglich.
	Die schwersten Schätze werden zuerst genommen.
    */
    func SchätzeTesten()
    {
        var gewicht = 0.0
        for akt in schätze
        {
            gewicht += akt.GewichtGeben()
        }
        if gewicht > maxGewicht
        {
            var immer = [Schatz]()
            var rest = [Schatz]()
            var opt = [Schatz]()
            var max = maxGewicht
            var pos = 0
            while pos < schätze.count
            {
                let akt = schätze[pos]
                if (akt.GewichtGeben() <= 0.0) || (akt is Schlüssel)
                {
                    immer.append(akt)
                    schätze.remove(at: pos)
                    max -= akt.GewichtGeben()
                }
                else
                {
                    pos = pos + 1
                }
            }
            if (handWaffe != nil) && (handWaffe!.GewichtGeben() <= max)
            {
                schätze.remove(at: schätze.firstIndex(where: {$0 === handWaffe})!)
                max -= handWaffe!.GewichtGeben()
                immer.append(handWaffe!)
            }
            if (schutz != nil) && (schutz!.GewichtGeben() <= max)
            {
                schätze.remove(at: schätze.firstIndex(where: {$0 === schutz})!)
                max -= schutz!.GewichtGeben()
                immer.append(schutz!)
            }
            while schätze.count > 0
            {
                var aktGewicht = 0.0
                var top: Schatz! = nil
                for akt in schätze
                {
                    if akt.GewichtGeben() > aktGewicht
                    {
                        top = akt
                        aktGewicht = akt.GewichtGeben()
                    }
                }
                if top.GewichtGeben() <= max
                {
                    opt.append(top)
                    max -= top.GewichtGeben()
                }
                else
                {
                    rest.append(top)
                }
                schätze.remove(at: schätze.firstIndex(where: {$0 === top})!)
            }
            schätze = opt
            for akt in immer
            {
                schätze.append(akt)
            }
            for akt in rest
            {
                wo.SchatzLagern(schatz: akt)
            }
        }
        if schätze.isEmpty
        {
            handWaffe = nil
            schutz = nil
        }
        else
        {
            if (handWaffe != nil) && !schätze.contains(where: {$0 === handWaffe!})
            {
                handWaffe = nil
            }
            if (schutz != nil) && !schätze.contains(where: {$0 === schutz!})
            {
                schutz = nil
            }
        }
    }

    /**
	Lässt den angegebenen Schatz da
	- parameters:
		- nummer: laufende Nummer des Schatzes
     */
    func SchatzDalassen(nummer: Int)
    {
        if nummer < schätze.count
        {
            let hilf = schätze.remove(at: nummer)
            wo.SchatzLagern(schatz: hilf)
            SchätzeTesten()
            InventarListeEintragen()
            RaumSchätzeEintragen()
        }
    }
    
    /**
	Nimmt den angegebenen Schatz mit
	- parameters:
		- nummer: laufende Nummer des Schatzes
     */
    func SchatzMitnehmen(nummer: Int)
    {
        
        if nummer < wo.schätze.count
        {
            var gewicht = 0.0
            var mentalGewicht = 0
            for akt in schätze
            {
                gewicht += akt.GewichtGeben()
                if akt is Spruch
                {
                    mentalGewicht += (akt as! Spruch).MentalgewichtGeben()
                }
            }
            let hilf = wo.schätze[nummer]
            if (gewicht + hilf.GewichtGeben() <= maxGewicht) && (hilf is Spruch ? mentalGewicht + (hilf as! Spruch).MentalgewichtGeben() <= 20 + erfahrung : true)
            {
                wo.schätze.remove(at: nummer)
                schätze.append(hilf)
                if (hilf is Spruch)
                {
                    (hilf as! Spruch).GefundenMelden()
                }
                InventarListeEintragen()
                RaumSchätzeEintragen()
            }
            else
            {
                spiel.BotschaftSenden(botschaft: "Soviel können Sie nicht mitnehmen, das Maximalgewicht ist: \(maxGewicht)" + (hilf is Spruch ? ", das maximale Mentalgewicht: \(20 + erfahrung)" : ""))
            }
        }
    }
    
    /**
    Trägt die Schätze des Raums in die Anzeige ein.
    */
    func RaumSchätzeEintragen()
    {
        var texte = [String]()
        var nummer = 0
        for akt in wo.schätze
        {
            texte.append("\(nummer) " + akt.NameGeben() + " \(akt.gewicht)")
            nummer = nummer + 1
        }
        spiel.anzeige.RaumSchätzeSetzen(texte: texte)
    }
    
    /**
    Trägt die Daten des Spielers in die Anzeige ein.
    */
    func InventarListeEintragen()
    {
        var gewicht = 0.0;
        var texte = [String]()
        for akt in schätze
        {
            gewicht += akt.GewichtGeben();
        }
        spiel.anzeige.SchlüsselAnzahlSetzen(anz: AnzahlSchlüsselGeben());
        spiel.anzeige.GesamtgewichtSetzen(gewicht: gewicht);
        var nummer = 0
        for akt in schätze
        {
            texte.append("\(nummer) " + akt.NameGeben() + " \(akt.gewicht)")
            nummer = nummer + 1
        }
        spiel.anzeige.SpielerSchätzeSetzen(texte: texte);
        if handWaffe == nil
        {
            spiel.anzeige.HandwaffenNameSetzen(name: "keine")
        }
        else
        {
            spiel.anzeige.HandwaffenNameSetzen(name: handWaffe!.NameGeben())
        }
        if schutz == nil
        {
            spiel.anzeige.SchutzwaffenNameSetzen(name: "keine")
        }
        else
        {
            spiel.anzeige.SchutzwaffenNameSetzen(name: schutz!.NameGeben())
        }
    }
    
    /**
    Aktiviert den Schatz gemäß der Anweisung
	- parameters:
		- nummer: der zu aktivierende Schatz
    */
    func SchatzAktivieren(nummer: Int)
    {
        let hilf = schätze[nummer]
        if hilf is AngriffsWaffe
        {
            handWaffe = (hilf as! AngriffsWaffe)
            spiel.anzeige.HandwaffenNameSetzen(name: handWaffe!.NameGeben())
        }
        else if hilf is SchutzWaffe
        {
            schutz = (hilf as! SchutzWaffe)
            spiel.anzeige.SchutzwaffenNameSetzen(name: schutz!.NameGeben())
        }
        else if (hilf is Spruch)
        {
            (hilf as! Spruch).Ausführen()
        }
    }
    
    /**
	Ändert die Stärke um den angegebenen Betrag.
	
	Die Grenzen für die Stärke werden sicher gestellt.
	- parameters:
		- wert: der Wert, um den sich die Stärke (höchstens) ändert
     */
    func StärkeÄndern(wert: Int)
    {
        stärke += wert
        if stärke < 0
        {
            stärke = 0
        }
        else if stärke > maxStärke
        {
            stärke = maxStärke
        }
        spiel.anzeige.StärkeSetzen(st: stärke)
    }

	func RaumBetreten (monsterZählt: Bool, besiegt: Bool) -> Bool
	{
        if monsterZählt && besiegt
        {
            erfahrung = erfahrung + 2
        }
        spiel.anzeige.KampfZuEndeSetzen()
        spiel.gewölbe.AnzeigeZeichnen(alle: false, markierung: wo)
        if spiel.gewölbe.AusgangTesten(test: wo) && (AnzahlSchlüsselGeben() == 3)
        {
            return true
        }
        if besiegt
        {
            SchätzeTesten()
            InventarListeEintragen()
            if !wo.SchatzVorhanden()
            {
                StärkeÄndern(wert: 1)
            }
            RaumSchätzeEintragen()
        }
        else
        {
            wo.monster.KampfAbbrechen(sp: self)
            SchätzeTesten()
            InventarListeEintragen()
        }
        return false
    }

    /**
	Bestimmt den nächsten Raum bzw. die Weiterführung eines Monsterkamps

	Vorher können Waffen und Spürüche gewählt werden.
	- returns: wahr, wenn Spielende errecht wurde
    */
    func KommandosAusführen(kdo: Kommandos, param: Int) -> Bool
    {
        switch kdo
        {
            case .zimmerWechsel:
                let raum = param
                spiel.BotschaftSenden(botschaft: "Sie gehen nach: \(raum)")
                if NachRaumGehen(nummer: raum)
                {
                    let monsterZählt = !(wo.monster is LeeresMonster)
                    wo.monster.KampfBeginnen(sp: self)
                    if monsterZählt
                    {
                        erfahrung = erfahrung + 1
                    }
                    if wo.monster.KampfDurchführen(sp: self)
                    {
                    	if RaumBetreten(monsterZählt: monsterZählt, besiegt: true)
                    	{
                    		return true
                        }
                    }
                    else
                    {
                        SchätzeTesten()
                        InventarListeEintragen()
                        spiel.anzeige.RundeVerlorenSetzen(text: wo.monster.NameGeben())
                    }
                    
                }
                else
                {
                    spiel.BotschaftSenden(botschaft: "Diesen Raum können Sie nicht erreichen")
                }
            case .weiterKämpfen:
                erfahrung = erfahrung + 1
                if wo.monster.KampfDurchführen(sp: self)
                {
                    if RaumBetreten(monsterZählt: true, besiegt: true)
                    {
                        return true
                    }
                }
                else
                {
                    SchätzeTesten()
                    InventarListeEintragen()
                }
            case .kampfAbbrechen:
                if RaumBetreten(monsterZählt: true, besiegt: false)
                {
                    return true
                }
            case .handwaffeWeglegen:
                handWaffe = nil
                spiel.anzeige.HandwaffenNameSetzen(name: "keine")
            case .schutzwaffeWeglegen:
                schutz = nil
                spiel.anzeige.SchutzwaffenNameSetzen(name: "keine")
            case .schatzAktivieren:
                SchatzAktivieren(nummer: param)
            case .schatzDalassen:
                SchatzDalassen(nummer: param)
            case .schatzMitnehmen:
                SchatzMitnehmen(nummer: param)
        }
        return false
    }
}

/**
Verwaltet das Spiel.

- author: Albert Wiedemann
- version: 1.0
*/
class Spiel
{
    /** Zugang zur Anzeige für das Spiel */
    let anzeige: AnzeigeInterface
    /** Der Spieler */
    var spieler: Spieler!
    /** Das Gewölbe */
    var gewölbe: Gewölbe!

    /**
	Legt die Grunddaten des Spiels an
	- parameters:
		- kontroller: Interface zum Kontroller für das Spiel
		- anzeige: Interface zur Anzeige für das Spiel
     */
    init(anzeige: AnzeigeInterface)
    {
        self.anzeige = anzeige
        spieler = nil
        gewölbe = nil
        gewölbe =  Gewölbe(spiel: self)
        gewölbe.Generieren()
        spieler = Spieler(spiel: self, start: gewölbe.StartRaum())
        spieler.StärkeÄndern(wert: 0)        /* Anfangsstärke setzen */
        gewölbe.AnzeigeZeichnen(alle: false, markierung: spieler.wo)
    }
    
    /**
	Gibt eine Botschaft an den Benutzer aus.
	- parameters:
		- botschaft: der Botschaftstext
    */
    func BotschaftSenden(botschaft: String)
    {
        anzeige.NeueBotschaftEintragen(text: botschaft)
    }
}


/**
Rahmenklasse des Spiels

- author: Albert Wiedemann
- version: 1.0
*/
class Adventure
{
	var a: Anzeige
    /**
    Legt die zentralen Objekte an und startet das Spiel
    */
    init()
    {
        let k = Kontroller()
        a = Anzeige(k: k)
        PlaygroundPage.current.liveView = a
        let s = Spiel(anzeige: a)
        k.SpielSetzen(s: s)
    }
}

Adventure()
